"""Utilities for handling streaming datasets."""
from .cache import Cache
from .array import iter_array
from .array import iter_pandas
from .array import iter_sklearn_dataset
from .csv import iter_csv
from .libsvm import iter_libsvm
from .qa import simulate_qa
from .shuffle import shuffle
from .vaex import iter_vaex


__all__ = [
    'Cache',
    'iter_array',
    'iter_csv',
    'iter_libsvm',
    'iter_pandas',
    'iter_sklearn_dataset',
    'iter_vaex',
    'simulate_qa',
    'shuffle'
]



import collections
import types

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from .. import base

from . import func


__all__ = ['TransformerUnion']


class TransformerUnion(collections.UserDict, base.Transformer):
    """Packs multiple transformers into a single one.

    Calling ``transform_one`` will concatenate each transformer's output using a
    `collections.ChainMap`.

    Parameters:
        transformers (list): transformers to pack together.

    Example:

        ::

            >>> from pprint import pprint
            >>> import creme.compose
            >>> import creme.feature_extraction
            >>> import creme.stats

            >>> X = [
            ...     {'place': 'Taco Bell', 'revenue': 42},
            ...     {'place': 'Burger King', 'revenue': 16},
            ...     {'place': 'Burger King', 'revenue': 24},
            ...     {'place': 'Taco Bell', 'revenue': 58},
            ...     {'place': 'Burger King', 'revenue': 20},
            ...     {'place': 'Taco Bell', 'revenue': 50}
            ... ]

            >>> mean = creme.feature_extraction.Agg(
            ...     on='revenue',
            ...     by='place',
            ...     how=creme.stats.Mean()
            ... )
            >>> count = creme.feature_extraction.Agg(
            ...     on='revenue',
            ...     by='place',
            ...     how=creme.stats.Count()
            ... )
            >>> agg = creme.compose.TransformerUnion(mean)
            >>> agg += count

            >>> for x in X:
            ...     pprint(agg.fit_one(x).transform_one(x))
            {'revenue_count_by_place': 1, 'revenue_mean_by_place': 42.0}
            {'revenue_count_by_place': 1, 'revenue_mean_by_place': 16.0}
            {'revenue_count_by_place': 2, 'revenue_mean_by_place': 20.0}
            {'revenue_count_by_place': 2, 'revenue_mean_by_place': 50.0}
            {'revenue_count_by_place': 3, 'revenue_mean_by_place': 20.0}
            {'revenue_count_by_place': 3, 'revenue_mean_by_place': 50.0}

            >>> pprint(agg.transform_one({'place': 'Taco Bell'}))
            {'revenue_count_by_place': 3, 'revenue_mean_by_place': 50.0}

    """

    def __init__(self, *transformers):
        super().__init__()
        for transformer in transformers:
            self += transformer

    @property
    def is_supervised(self):
        return any(transformer.is_supervised for transformer in self.values())

    def __str__(self):
        return ' + '.join(map(str, self.values()))

    def __repr__(self):
        return (
            'TransformerUnion (\n\t' +
            '\t'.join(',\n'.join(map(repr, self.values())).splitlines(True)) +
            '\n)'
        ).expandtabs(2)

    def _get_params(self):
        return dict(self.items())

    def _set_params(self, new_params=None):
        if new_params is None:
            new_params = {}
        return TransformerUnion(*[
            (name, new_params[name])
            if isinstance(new_params.get(name), base.Estimator) else
            (name, step._set_params(new_params.get(name, {})))
            for name, step in self.items()
        ])

    def add_step(self, transformer):
        """Adds a transformer while taking care of the input type."""

        name = None
        if isinstance(transformer, tuple):
            name, transformer = transformer

        # If the step is a function then wrap it in a FuncTransformer
        if isinstance(transformer, (types.FunctionType, types.LambdaType)):
            transformer = func.FuncTransformer(transformer)

        def infer_name(transformer):
            if isinstance(transformer, func.FuncTransformer):
                return infer_name(transformer.func)
            elif isinstance(transformer, (types.FunctionType, types.LambdaType)):
                return transformer.__name__
            elif hasattr(transformer, '__class__'):
                return transformer.__class__.__name__
            return str(transformer)

        # Infer a name if none is given
        if name is None:
            name = infer_name(transformer)

        if name in self:
            counter = 1
            while f'{name}{counter}' in self:
                counter += 1
            name = f'{name}{counter}'

        # Store the transformer
        self[name] = transformer

        return self

    def __add__(self, other):
        return self.add_step(other)

    def fit_one(self, x, y=None):
        for transformer in self.values():
            transformer.fit_one(x, y)
        return self

    def transform_one(self, x):
        """Passes the data through each transformer and packs the results together."""
        return dict(collections.ChainMap(*(
            transformer.transform_one(x)
            for transformer in self.values()
        )))

    def draw(self):

        if not GRAPHVIZ_INSTALLED:
            raise ImportError('graphviz is not installed')

        g = graphviz.Digraph(engine='fdp')

        for part in self.values():
            if hasattr(part, 'draw'):
                g.subgraph(part.draw())
            else:
                g.node(str(part))
        return g



"""Loss functions."""
from libc cimport math
from .. import utils


__all__ = [
    'Absolute',
    'BinaryFocalLoss',
    'Cauchy',
    'CrossEntropy',
    'Hinge',
    'EpsilonInsensitiveHinge',
    'Log',
    'Perceptron',
    'Poisson',
    'Quantile',
    'Squared'
]


cdef double clamp_proba(double x):
    return math.fmax(math.fmin(x, 1 - 1e-15), 1e-15)


cdef class Loss:

    def __str__(self):
        return utils.pretty.format_object(self)

    def mean_func(self, y_pred):
        """Mean function.

        This is the inverse of the link function.

        References:
            1. `Wikipedia section on link and mean function <https://www.wikiwand.com/en/Generalized_linear_model#/Link_function>`_

        """


cdef class ClassificationLoss(Loss):
    pass


cdef class BinaryLoss(ClassificationLoss):
    """A loss appropriate for binary classification tasks."""

    cpdef double eval(self, bint y_true, double y_pred):
        """Returns the loss."""

    cpdef double gradient(self, bint y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``."""

    cpdef double mean_func(self, double y_pred):
        return utils.math.sigmoid(y_pred)


cdef class MultiClassLoss(ClassificationLoss):
    """A loss appropriate for multi-class classification tasks."""

    cpdef double eval(self, object y_true, dict y_pred):
        """Returns the loss."""

    cpdef dict gradient(self, object y_true, dict y_pred):
        """Returns the gradient with respect to ``y_pred``."""

    cpdef dict mean_func(self, dict y_pred):
        return utils.math.softmax(y_pred)


cdef class RegressionLoss(Loss):
    """A loss appropriate for regression tasks."""

    cpdef double eval(self, double y_true, double y_pred):
        """Returns the loss."""

    cpdef double gradient(self, double y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``."""

    cpdef double mean_func(self, double y_pred):
        return y_pred


cdef class Absolute(RegressionLoss):
    """Absolute loss, also known as the mean absolute error or L1 loss.

    Mathematically, it is defined as

    .. math:: L = |p_i - y_i|

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = sgn(p_i - y_i)

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Absolute()
            >>> loss.eval(-42, 42)
            84.0
            >>> loss.gradient(1, 2)
            1.0
            >>> loss.gradient(2, 1)
            -1.0

    """

    cpdef double eval(self, double y_true, double y_pred):
        return abs(y_pred - y_true)

    cpdef double gradient(self, double y_true, double y_pred):
        if y_pred > y_true:
            return 1
        return -1


cdef class Cauchy(RegressionLoss):
    """Cauchy loss function.

    References:
        1. `"Effect of MAE" Kaggle discussion <https://www.kaggle.com/c/allstate-claims-severity/discussion/24520#140163>`_
        2. `Paris Madness Kaggle kernel <https://www.kaggle.com/raddar/paris-madness>`_

    """

    def __init__(self, C=80):
        self.C = C

    cpdef double eval(self, double y_true, double y_pred):
        return math.fabs(y_pred - y_true)

    cpdef double gradient(self, double y_true, double y_pred):
        diff = y_pred - y_true
        return diff / ((diff / self.C) ** 2 + 1)


cdef class CrossEntropy(MultiClassLoss):
    """Cross entropy loss.

    This is a generalization of logistic loss to multiple classes.

    Example:

        ::

            >>> from creme import optim

            >>> y_true = [0, 1, 2, 2]
            >>> y_pred = [
            ...     {0: 0.29450637, 1: 0.34216758, 2: 0.36332605},
            ...     {0: 0.21290077, 1: 0.32728332, 2: 0.45981591},
            ...     {0: 0.42860913, 1: 0.33380113, 2: 0.23758974},
            ...     {0: 0.44941979, 1: 0.32962558, 2: 0.22095463}
            ... ]

            >>> loss = optim.losses.CrossEntropy()

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(loss.eval(y_t, y_p))
            1.222454
            1.116929
            1.437209
            1.509797

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(loss.gradient(y_t, y_p))
            {0: -0.70549363, 1: 0.34216758, 2: 0.36332605}
            {0: 0.21290077, 1: -0.67271668, 2: 0.45981591}
            {0: 0.42860913, 1: 0.33380113, 2: -0.76241026}
            {0: 0.44941979, 1: 0.32962558, 2: -0.77904537}

    References:
        1. `What is Softmax regression and how is it related to Logistic regression? <https://github.com/rasbt/python-machine-learning-book/blob/master/faq/softmax_regression.md>`_

    """

    cdef readonly dict class_weight

    def __init__(self, class_weight=None):
        if class_weight is None:
            class_weight = {}
        self.class_weight = class_weight

    cpdef double eval(self, object y_true, dict y_pred):
        cdef double total = 0

        for label, proba in y_pred.items():
            if y_true == label:
                total += self.class_weight.get(label, 1.) * math.log(clamp_proba(proba))

        return -total

    cpdef dict gradient(self, object y_true, dict y_pred):
        return {
            label: (
                self.class_weight.get(label, 1.) *
                (clamp_proba(y_pred.get(label, 0.)) - (y_true == label))
            )
            for label in {*y_pred.keys(), y_true}
        }


cdef class Hinge(BinaryLoss):
    """Computes the hinge loss.

    Mathematically, it is defined as

    .. math:: L = max(0, 1 - p_i * y_i)

    It's gradient w.r.t. to $p_i$ is

    .. math::
        \\frac{\\partial L}{\\partial y_i} = \\left\{
        \\begin{array}{ll}
            \\ 0  &   p_iy_i \geqslant 1  \\\\
            \\ - y_i & p_iy_i < 1
        \\end{array}
        \\right.

    Parameters:
        threshold (float): Margin threshold. 1 yield the loss used in SVMs, whilst 0 is equivalent to
            the loss used in the Perceptron algorithm.

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Hinge(threshold=1)
            >>> loss.eval(1, .2)
            0.8

            >>> loss.gradient(1, .2)
            -0.2

    """

    cdef readonly double threshold

    def __init__(self, double threshold=1.):
        self.threshold = threshold

    cpdef double eval(self, bint y_true, double y_pred):
        # Convert 0 to -1
        y_true = y_true or -1
        return math.fmax(0, self.threshold - y_true * y_pred)

    cpdef double gradient(self, bint y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``.

        References:
            1. `WolframAlpha derivation <https://www.wolframalpha.com/input/?i=derivative+max(0,+1+-+p+*+y)+wrt+p>`_

        """
        y_true = y_true or -1
        if y_true * y_pred < self.threshold:
            return -y_pred
        return 0


cdef class EpsilonInsensitiveHinge(RegressionLoss):

    cdef readonly double eps

    def __init__(self, eps=0.1):
        self.eps = eps

    cpdef double eval(self, double y_true, double y_pred):
        # Convert 0 to -1
        y_true = y_true or -1
        return math.fmax(0, math.fabs(y_pred - y_true) - self.eps)

    cpdef double gradient(self, double y_true, double y_pred):
        """Returns the gradient with respect to ``y_pred``.

        References:
            1. `WolframAlpha derivation <https://www.wolframalpha.com/input/?i=derivative+max(0,+abs(p+-+y)+-+eps)+wrt+p>`_

        """
        y_true = y_true or -1
        if y_pred > y_true + self.eps:
            return 1
        elif y_pred + self.eps < y_true:
            return -1
        return 0


cdef class Log(BinaryLoss):
    """Logarithmic loss.

    This loss function expects each provided ``y_pred`` to be a logit. In other words if must be
    the raw output of a linear model or a neural network.

    References:
        1. `Logit Wikipedia page <https://www.wikiwand.com/en/Logit>`_

    """

    cdef readonly double weight_pos
    cdef readonly double weight_neg

    def __init__(self, weight_pos=1., weight_neg=1.):
        self.weight_pos = weight_pos
        self.weight_neg = weight_neg

    cpdef double eval(self, bint y_true, double y_pred):
        weight = self.weight_pos
        if y_true == 0:
            y_true = -1
            weight = self.weight_neg

        z = y_pred * y_true
        if z > 18.:
            return weight * math.exp(-z)
        if z < -18.:
            return weight * -z
        return weight * math.log(1. + math.exp(-z))

    cpdef double gradient(self, bint y_true, double y_pred):
        weight = self.weight_pos
        if y_true == 0:
            y_true = -1
            weight = self.weight_neg

        z = y_pred * y_true
        if z > 18.:
            return weight * math.exp(-z) * -y_true
        if z < -18.:
            return weight * -y_true
        return weight * -y_true / (math.exp(z) + 1.)


cdef class Quantile(RegressionLoss):
    """Quantile loss.

    Parameters:
        alpha (float): Desired quantile to attain.

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Quantile(0.5)
            >>> loss.eval(1, 3)
            1.0

            >>> loss.gradient(1, 3)
            0.5

            >>> loss.gradient(3, 1)
            -0.5

    References:
        1. `Wikipedia article on quantile regression <https://www.wikiwand.com/en/Quantile_regression>`_
        2. `Derivative from WolframAlpha <https://www.wolframalpha.com/input/?i=derivative+(y+-+p)+*+(alpha+-+Boole(y+-+p))+wrt+p>`_

    """

    cdef readonly double alpha

    def __init__(self, alpha):
        self.alpha = alpha

    cpdef double eval(self, double y_true, double y_pred):
        diff = y_pred - y_true
        return (self.alpha - (diff < 0)) * diff

    cpdef double gradient(self, double y_true, double y_pred):
        return (y_true < y_pred) - self.alpha


cdef class Squared(RegressionLoss):
    """Squared loss, also known as the L2 loss.

    Mathematically, it is defined as

    .. math:: L = (p_i - y_i) ^ 2

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = 2 \times (p_i - y_i)

    One thing to note is that this convention is consistent with Vowpal Wabbit and PyTorch, but
    not with scikit-learn. Indeed scikit-learn divides the loss by 2, making the 2 dissapear in
    the gradient.

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.losses.Squared()
            >>> loss.eval(-4, 5)
            81.0
            >>> loss.gradient(-4, 5)
            18.0
            >>> loss.gradient(5, -4)
            -18.0

    """

    cpdef double eval(self, double y_true, double y_pred):
        return (y_pred - y_true) * (y_pred - y_true)

    cpdef double gradient(self, double y_true, double y_pred):
        return 2. * (y_pred - y_true)


class BinaryFocalLoss(BinaryLoss):
    """Binary focal loss.

    This implements the "star" algorithm from the appendix of the focal loss paper.

    Parameters:
        gamma (float)
        beta (float)

    Refenrences:
        1. `Lin, T.Y., Goyal, P., Girshick, R., He, K. and Dollár, P., 2017. Focal loss for dense object detection. In Proceedings of the IEEE international conference on computer vision (pp. 2980-2988). <https://arxiv.org/pdf/1708.02002.pdf>`_

    """

    def __init__(self, gamma=2, beta=1):
        self.gamma = gamma
        self.beta = beta

    def eval(self, y_true, y_pred):

        # Convert 0 to -1
        y_true = int(y_true or -1)

        xt = y_true * y_pred
        pt = utils.math.sigmoid(self.gamma * xt + self.beta)

        return -math.log(pt) / self.gamma

    def gradient(self, y_true, y_pred):

        # Convert 0 to -1
        y_true = int(y_true or -1)

        xt = y_true * y_pred
        pt = utils.math.sigmoid(self.gamma * xt + self.beta)

        return y_true * (pt - 1)


cdef class Poisson(RegressionLoss):
    """Poisson loss.

    The Poisson loss is usually more suited for regression with count data than the squared loss.

    Mathematically, it is defined as

    .. math:: L = exp(p_i) - y_i \\times p_i

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = exp(p_i) - y_i

    """

    cpdef double eval(self, double y_true, double y_pred):
        return math.exp(y_pred) - y_true * y_pred

    cpdef double gradient(self, double y_true, double y_pred):
        return math.exp(y_pred) - y_true

    cpdef double mean_func(self, double y_pred):
        return math.exp(y_pred)


cdef class Perceptron(Hinge):
    """Perceptron loss.

    The Perceptron loss is the loss used in the Perceptron algorithm. Using this loss in a logistic
    regression yields the Perceptron algorithm.

    Mathematically, it is defined as

    .. math:: L = exp(p_i) - y_i \\times p_i

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = exp(p_i) - y_i

    References:
        1. `Wikipedia page on the Perceptron algorithm <https://www.wikiwand.com/en/Perceptron>`_

    """

    def __init__(self):
        super().__init__(threshold=0.)



import abc

from . import estimator


class Transformer(estimator.Estimator):
    """A transformer."""

    def fit_one(self, x: dict, y=None) -> 'Transformer':
        """Fits to a set of features ``x`` and an optional target ``y``.

        A lot of transformers don't actually have to do anything during the ``fit_one`` step
        because they are stateless. For this reason the default behavior of this function is to do
        nothing. Transformers that however do something during the ``fit_one`` can override this
        method.

        Parameters:
            x (dict)
            y (optional)

        Returns:
            self

        """
        return self

    @abc.abstractmethod
    def transform_one(self, x: dict) -> dict:
        """Transforms a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            dict

        """

    @property
    def is_supervised(self) -> bool:
        """Indicates if the transformer uses the target ``y`` or not.

        Supervised transformers have to be handled differently from unsupervised transformers in an
        online setting. This is especially true for target encoding where leakage can easily occur.
        Most transformers are unsupervised and so this property returns by default ``False``.
        Transformers that are supervised must override this property in their definition.

        """
        return False

    def __or__(self, other):
        """Merges with another Transformer into a Pipeline."""
        from .. import compose
        if isinstance(other, compose.Pipeline):
            return other.__ror__(self)
        return compose.Pipeline(self, other)

    def __ror__(self, other):
        """Merges with another Transformer into a Pipeline."""
        from .. import compose
        if isinstance(other, compose.Pipeline):
            return other.__or__(self)
        return compose.Pipeline(other, self)

    def __add__(self, other):
        """Merges with another Transformer into a TransformerUnion."""
        from .. import compose
        if isinstance(other, compose.TransformerUnion):
            return other.__add__(self)
        return compose.TransformerUnion(self, other)

    def __radd__(self, other):
        """Merges with another Transformer into a TransformerUnion."""
        from .. import compose
        if isinstance(other, compose.TransformerUnion):
            return other.__add__(self)
        return compose.TransformerUnion(other, self)



import os

from .. import stream

from . import base


class Airline(base.FileDataset):
    """Monthly number of international airline passengers.

    The stream contains 144 items and only one single feature, which is the month. The goal is to
    predict the number of passengers each month by capturing the trend and the seasonality of the
    data.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `International airline passengers: monthly totals in thousands. Jan 49 – Dec 60 <https://datamarket.com/data/set/22u3/international-airline-passengers-monthly-totals-in-thousands-jan-49-dec-60#!ds=22u3&display=line>`_

    """

    def __init__(self):
        super().__init__(
            n_samples=144,
            n_features=1,
            category=base.REG
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            os.path.join(directory, 'airline-passengers.csv'),
            target_name='passengers',
            converters={'passengers': int},
            parse_dates={'month': '%Y-%m'}
        )



import glob
import os

from . import base


class MaliciousURL(base.FileDataset):
    """Malicious URLs dataset.

    This dataset contains features about URLs that are classified as malicious or not.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Detecting Malicious URLs <http://www.sysnet.ucsd.edu/projects/url/>`_
        2. `Identifying Suspicious URLs: An Application of Large-Scale Online Learning <http://cseweb.ucsd.edu/~jtma/papers/url-icml2009.pdf>`_

    """

    def __init__(self, data_home=None, verbose=True):
        super().__init__(
            n_samples=2_396_130,
            n_features=3_231_961,
            category=base.BINARY_CLF,
            url='http://www.sysnet.ucsd.edu/projects/url/url_svmlight.tar.gz',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):

        files = glob.glob(f'{directory}/url_svmlight/Day*.svm')
        files.sort(key=lambda x: int(os.path.basename(x).split('.')[0][3:]))

        def parse_libsvm_feature(f):
            k, v = f.split(':')
            return int(k), float(v)

        # There are 150 files, where each file corresponds to a day
        for file in files:
            with open(file) as f:
                for line in f:
                    # Each file has the libsvm format
                    elements = line.rstrip().split(' ')
                    y = elements.pop(0) == '+1'
                    x = dict(parse_libsvm_feature(f) for f in elements)
                    yield x, y



import importlib
import inspect
import urllib.request

import pytest


def _iter_datasets():
    return inspect.getmembers(importlib.import_module(f'creme.datasets'), inspect.isclass)


@pytest.mark.parametrize(
    'dataset',
    [
        pytest.param(dataset(), id=name)
        for name, dataset
        in _iter_datasets()
        if dataset()._remote
    ]
)
@pytest.mark.web
def test_remote_url_200(dataset):
    r = urllib.request.urlopen(dataset.dl_params['url'])
    assert r.getcode() == 200


@pytest.mark.parametrize(
    'dataset',
    [
        pytest.param(dataset(), id=name)
        for name, dataset
        in _iter_datasets()
    ]
)
@pytest.mark.web
def test_size(dataset):
    n = 0
    for x, _ in dataset:
        assert len(x) == dataset.n_features
        n += 1
    assert n == dataset.n_samples



from .. import stream

from . import base


class TREC07(base.FileDataset):
    """TREC's 2007 Spam Track dataset.

    The data contains 75,419 chronologically ordered items, i.e. 3 months of emails delivered
    to a particular server in 2007. Spam messages represent 66.6% of the dataset.
    The goal is to predict whether an email is a spam or not.

    The available raw features are: sender, recipients, date, subject, body.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `TREC 2007 Spam Track Overview <https://trec.nist.gov/pubs/trec16/papers/SPAM.OVERVIEW16.pdf>`_
        2. `Code ran to parse the dataset <https://gist.github.com/gbolmier/b6a942699aaaedec54041a32e4f34d40>`_

    """

    def __init__(self, data_home=None, verbose=True):
        super().__init__(
            n_samples=75_419,
            n_features=5,
            category=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/trec07p.zip',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            f'{directory}/trec07p.csv',
            target_name='y',
            delimiter=',',
            quotechar='"',
            field_size_limit=1_000_000,
        )



def iter_vaex(X, y=None, features=None, **kwargs):
    """Yields rows from a ``vaex.DataFrame``.

    Parameters:
        X (vaex.DataFrame): A vaex DataFrame housing the training featuers.
        y (string or vaex.Expression): The column or expression containing the target variable.
        features (list of strings or vaex.Expressions): A list of features used for training.
        If None, all columns in ``X`` will be used. Features specifying in ``y`` are ignored.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """

    from vaex.utils import _ensure_strings_from_expressions, _ensure_list

    features = _ensure_strings_from_expressions(features)
    feature_names = features or X.get_column_names()

    if y:
        y = _ensure_strings_from_expressions(y)
        y = _ensure_list(y)
        feature_names = [feat for feat in feature_names if feat not in y]

    multioutput = len(y) > 1

    if multioutput:
        for i in range(len(X)):
            yield (
                {key: X.evaluate(key, i, i + 1)[0] for key in feature_names},
                {key: X.evaluate(key, i, i + 1)[0] for key in y}
            )

    else:

        for i in range(len(X)):
            yield (
                {key: X.evaluate(key, i, i + 1)[0] for key in feature_names},
                X.evaluate(y[0], i, i + 1)[0]
            )



import collections
import functools
import operator
import random

from .. import base
from .. import preprocessing
from ..tree.base import Leaf, Branch, Split


__all__ = ['HalfSpaceTrees']


def make_padded_tree(limits, height, padding, rng=random, **node_params):

    if height == 0:
        return Leaf(**node_params)

    # Randomly pick a feature
    # We weight each feature by the gap between each feature's limits
    on = rng.choices(
        population=list(limits.keys()),
        weights=[limits[i][1] - limits[i][0] for i in limits]
    )[0]

    # Pick a split point; use padding to avoid too narrow a split
    a = limits[on][0]
    b = limits[on][1]
    at = rng.uniform(a + padding * (b - a), b - padding * (b - a))

    # Build the left node
    tmp = limits[on]
    limits[on] = (tmp[0], at)
    left = make_padded_tree(limits=limits, height=height - 1, padding=padding, rng=rng, **node_params)
    limits[on] = tmp

    # Build the right node
    tmp = limits[on]
    limits[on] = (at, tmp[1])
    right = make_padded_tree(limits=limits, height=height - 1, padding=padding, rng=rng, **node_params)
    limits[on] = tmp

    split = Split(on=on, how=operator.lt, at=at)
    return Branch(split=split, left=left, right=right, **node_params)


class HalfSpaceTrees(base.AnomalyDetector):
    """Half-Space Trees (HST).

    Half-space trees are an online variant of isolation forests. They work well when anomalies are
    spread out in time. They do not work well if anomalies are packed together in windows of time,
    such as in the KDD'99 HTTP and SMTP datasets.

    Parameters:
        n_trees (int): Number of trees to use.
        height (int): Height of each tree. Note that a tree of height ``h`` is made up of ``h + 1``
            levels and therefore contains ``2 ** (h + 1) - 1`` nodes.
        window_size (int): Number of observations to use for calculating the mass at each node in
            each tree.
        limits (dict): Specifies the range of each feature. By default each feature is assumed to
            be in range ``[0, 1]``.
        seed (int): Random number seed.

    Example:

        ::

            >>> from creme import anomaly

            >>> X = [0.5, 0.45, 0.43, 0.44, 0.445, 0.45, 0.0]
            >>> hst = anomaly.HalfSpaceTrees(
            ...     n_trees=5,
            ...     height=3,
            ...     window_size=3,
            ...     seed=42
            ... )

            >>> for x in X[:3]:
            ...     hst = hst.fit_one({'x': x})  # Warming up

            >>> for x in X:
            ...     features = {'x': x}
            ...     hst = hst.fit_one(features)
            ...     print(f'Anomaly score for x={x:.3f}: {hst.score_one(features):.3f}')
            Anomaly score for x=0.500: 0.107
            Anomaly score for x=0.450: 0.071
            Anomaly score for x=0.430: 0.107
            Anomaly score for x=0.440: 0.107
            Anomaly score for x=0.445: 0.107
            Anomaly score for x=0.450: 0.071
            Anomaly score for x=0.000: 0.853

    References:
        1. `Tan, S.C., Ting, K.M. and Liu, T.F., 2011, June. Fast anomaly detection for streaming data. In Twenty-Second International Joint Conference on Artificial Intelligence. <https://www.ijcai.org/Proceedings/11/Papers/254.pdf>`_

    """

    def __init__(self, n_trees=25, height=15, window_size=250, limits=None, seed=None):
        self.n_trees = n_trees
        self.window_size = window_size
        self.height = height
        self.limits = collections.defaultdict(functools.partial(tuple, (0, 1)))
        if limits is not None:
            self.limits.update(limits)
        self.seed = seed
        self.rng = random.Random(seed)

        self.trees = []
        self.counter = 0
        self._first_window = True

    @property
    def size_limit(self):
        """This is the threshold under which the node search stops during the scoring phase.

        The value .1 is a magic constant indicated in the original paper.

        """
        return .1 * self.window_size

    @property
    def _max_score(self):
        """The largest potential anomaly score."""
        return self.n_trees * self.window_size * (2 ** (self.height + 1) - 1)

    def fit_one(self, x):

        # The trees are built when the first observation comes in
        if not self.trees:
            self.trees = [
                make_padded_tree(
                    limits={i: self.limits[i] for i in x},
                    height=self.height,
                    padding=.15,
                    rng=self.rng,
                    # kwargs
                    r_mass=0,
                    l_mass=0
                )
                for _ in range(self.n_trees)
            ]

        # Update each tree
        for tree in self.trees:
            for node in tree.path(x):
                node.l_mass += 1

        # Pivot the masses if necessary
        self.counter += 1
        if self.counter == self.window_size:
            for tree in self.trees:
                for node, _ in tree.iter_dfs():
                    node.r_mass = node.l_mass
                    node.l_mass = 0
            self._first_window = False
            self.counter = 0

        return self

    def score_one(self, x):

        if self._first_window:
            return 0

        score = 0.
        for tree in self.trees:
            for depth, node in enumerate(tree.path(x)):
                score += node.r_mass * 2 ** depth
                if node.r_mass < self.size_limit:
                    break

        # Normalize the score between 0 and 1
        score /= self._max_score

        # We want high score -> anomaly, but we have high score -> normal
        return 1 - score



"""Anomaly detection.

The estimators in ``anomaly`` are slightly different than the rest of the estimators. Instead of a
``predict_one`` method, each anomaly detector has a ``score_one`` method which returns an anomaly
score for a given set of features. High scores indicate anomalies whereas low scores indicate
normal observations will have low scores. Note that the range of the scores depends on each
particular estimator.

"""
from .hst import HalfSpaceTrees


__all__ = ['HalfSpaceTrees']



"""Generic branch and leaf implementation."""
import collections


class Split:
    """A data class for storing split details."""

    def __init__(self, on, how, at):
        self.on = on
        self.how = how
        self.at = at

    def __call__(self, x):
        return self.how(x[self.on], self.at)

    def __repr__(self):
        at = self.at
        if isinstance(at, float):
            at = f'{at:.5f}'
        return f'{self.on} {repr(self.how)} {at}'


class Node:

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class Branch(Node):

    def __init__(self, split, left, right, **kwargs):
        super().__init__(**kwargs)
        self.split = split
        self.left = left
        self.right = right

    def next(self, x):
        """Returns the next node where x belongs."""
        return self.left if self.split(x) else self.right

    def path(self, x):
        """Iterates over the nodes that lead to the leaf which contains x."""
        yield self
        yield from self.next(x).path(x)

    @property
    def size(self):
        """Number of descendants, including thyself."""
        return 1 + self.left.size + self.right.size

    @property
    def height(self):
        """Distance to the deepest node."""
        return 1 + max(self.left.height, self.right.height)

    def iter_dfs(self, depth=0):
        """Iterates over nodes in a depth-first manner.

        Example:

            >>> tree = Branch(
            ...     None,
            ...     Branch(
            ...         None,
            ...         Leaf(no=2),
            ...         Leaf(no=3),
            ...         no=1
            ...     ),
            ...     Leaf(no=4),
            ...     no=0
            ... )

            >>> for node, depth in tree.iter_dfs():
            ...     print(f'#{node.no}, depth {depth}')
            #0, depth 0
            #1, depth 1
            #2, depth 2
            #3, depth 2
            #4, depth 1

        """
        yield self, depth
        yield from self.left.iter_dfs(depth=depth + 1)
        yield from self.right.iter_dfs(depth=depth + 1)

    def iter_edges(self):
        """Iterates over edges in a depth-first manner.

        Example:

            >>> tree = Branch(
            ...     None,
            ...     Branch(
            ...         None,
            ...         Leaf(no=2),
            ...         Leaf(no=3),
            ...         no=1
            ...     ),
            ...     Leaf(no=4),
            ...     no=0
            ... )

            >>> for parent_no, child_no, parent, child, child_depth in tree.iter_edges():
            ...     print(parent_no, child_no, child_depth)
            None 0 0
            0 1 1
            1 2 2
            1 3 2
            0 4 1

        """
        counter = 0

        def iterate(node, depth):

            nonlocal counter
            no = counter

            if isinstance(node, Branch):
                for child in (node.left, node.right):
                    counter += 1
                    yield no, counter, node, child, depth + 1
                    if isinstance(child, Branch):
                        yield from iterate(child, depth=depth + 1)

        yield None, 0, None, self, 0
        yield from iterate(self, depth=0)


class Leaf(Node):

    def path(self, x):
        yield self

    @property
    def size(self):
        return 1

    @property
    def height(self):
        return 0

    def iter_dfs(self, depth=0):
        yield self, depth

    def iter_edges(self):
        yield None, 0, None, self, 0


def iter_blocks(tree, limits, depth=-1):
    """Returns the block which encloses each node at a given depth.
    This only makes sense if the branches of the provided tree use the ``<`` operator as a split
    rule.

    Parameters:
        tree (Node)
        limits (dict)
        depth (int): Desired tree depth. Set to ``-1`` to iterate over the leaves.

    Example:

        >>> import operator
        >>> tree = Branch(
        ...     Split('x', operator.lt, .5),
        ...     Leaf(no=0),
        ...     Branch(
        ...         Split('y', operator.lt, .5),
        ...         Leaf(no=1),
        ...         Leaf(no=2)
        ...     )
        ... )

        >>> for leaf, block in iter_blocks(tree, limits={'x': (0, 1), 'y': (0, 1)}):
        ...     print(leaf.no, block)
        0 {'x': (0, 0.5), 'y': (0, 1)}
        1 {'x': (0.5, 1), 'y': (0, 0.5)}
        2 {'x': (0.5, 1), 'y': (0.5, 1)}

    """
    if depth == 0 or isinstance(tree, Leaf):
        yield (tree, limits)
        return

    on, at = tree.split.on, tree.split.at

    l_limits = {**limits, on: (limits[on][0], at)} if on in limits else limits
    yield from iter_blocks(tree=tree.left, limits=l_limits, depth=depth - 1)

    r_limits = {**limits, on: (at, limits[on][1])} if on in limits else limits
    yield from iter_blocks(tree=tree.right, limits=r_limits, depth=depth - 1)



import os

from .. import stream

from . import base


class ImageSegments(base.FileDataset):
    """Image segments classification.

    This dataset contains features that describe image segments into 7 classes: brickface, sky,
    foliage, cement, window, path, and grass.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `UCI page <https://archive.ics.uci.edu/ml/datasets/Statlog+(Image+Segmentation)>`_

    """

    def __init__(self):
        super().__init__(
            n_samples=2310,
            n_features=18,
            category=base.MULTI_CLF
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            os.path.join(directory, 'segment.csv.zip'),
            target_name='category',
            converters={
                'region-centroid-col': int,
                'region-centroid-row': int,
                'short-line-density-5': float,
                'short-line-density-2': float,
                'vedge-mean': float,
                'vegde-sd': float,
                'hedge-mean': float,
                'hedge-sd': float,
                'intensity-mean': float,
                'rawred-mean': float,
                'rawblue-mean': float,
                'rawgreen-mean': float,
                'exred-mean': float,
                'exblue-mean': float,
                'exgreen-mean': float,
                'value-mean': float,
                'saturation-mean': float,
                'hue-mean': float
            }
        )



import collections

import random

from .. import base
from .. import optim
from .. import utils


class Triplet(collections.namedtuple('Triplet', 'x y loss')):

    def __lt__(self, other):
        return self.loss < other.loss


class HardSampling(base.Wrapper):
    """Hard sampler."""

    def __init__(self, model, loss_function, size, p,seed = None):
        self.model = model
        self.loss_function = loss_function
        self.pred_func = model.predict_one

        if isinstance(model, base.Classifier):
            self.pred_func = model.predict_proba_one

        elif isinstance(model, base.BinaryClassifier):
            self.pred_func = lambda x: model.predict_proba_one(x)[True]

        self.p = p
        self.size = size
        self.buffer = utils.SortedWindow(self.size)
        self.seed = seed
        self._rng = random.Random(seed)

    @property
    def _model(self):
        return self.model

    def predict_one(self, x):
        return self.model.predict_one(x)

    def fit_one(self, x, y):
        """
        Compute the loss of the input observation. If the loss is high enough,  the tuple (x, y)
        will be stored in the buffer. Update the model with observations stored in the buffer with a
        probability p. Update the model with input observation with a probability (1 - p). Compute
        the loss a second times if the model fitted an observation from the buffer.

        Parameters:
            x (Dict): Features.
            y (float): Target.

        """
        loss = self.loss_function.eval(y_true=y, y_pred=self.pred_func(x))

        if len(self.buffer) < self.size:
            self.buffer.append(Triplet(x=x, y=y, loss=loss))

        elif loss > self.buffer[0].loss:
            self.buffer.pop(0)

            self.buffer.append(Triplet(x=x, y=y, loss=loss))

        # Probability p
        if self._rng.uniform(0, 1) <= self.p:
            i = self._rng.randint(0, len(self.buffer) - 1)

            triplet = self.buffer.pop(i)

            self.model.fit_one(triplet.x, triplet.y)

            loss = self.loss_function.eval(y_true=triplet.y, y_pred=self.pred_func(triplet.x))

            self.buffer.append(Triplet(x=triplet.x, y=triplet.y, loss=loss))

        # Probability (1 - p).
        else:
            self.model.fit_one(x, y)

        return self


class HardSamplingRegressor(HardSampling):
    """Hard-sampling regressor.

    This wrapper enables a model to retrain on past samples who's output was hard to predict.
    This works by storing the hardest samples in a buffer of a fixed size. When a new sample
    arrives, the wrapped model is either trained on one of the buffered samples with a probability p
    or on the new sample with a probability (1 - p).

    The hardness of an observation is evaluated with a loss function that compares the sample's
    ground truth with the wrapped model's prediction. If the buffer is not full, then the sample
    is added to the buffer. If the buffer is full and the new sample has a bigger loss than the
    lowest loss in the buffer, then the sample takes it's place.

    Parameters:
        regressor (base.Regressor)
        loss_function (creme.optim.losses.RegressionLoss): Criterion used to evaluate the hardness of a sample.
        size (int): Size of the buffer.
        p (float): Probability of updating the model with a sample from the buffer instead of a
            new incoming sample.
        seed (int): Random seed.

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import sampling

            >>> model = preprocessing.StandardScaler()

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     sampling.HardSamplingRegressor(
            ...         regressor = linear_model.LinearRegression(),
            ...          p = 0.2,
            ...          size = 30,
            ...         seed = 42,
            ...     )
            ... )

            >>> model_selection.progressive_val_score(
            ...     datasets.TrumpApproval(),
            ...     model,
            ...     metrics.MAE(),
            ...     print_every = 500
            ... )
            [500] MAE: 3.240558
            [1,000] MAE: 1.949106
            MAE: 1.947374

    """
    def __init__(self, regressor, size, p, loss_function = None, seed=None):
        if loss_function is None:
            loss_function =  optim.losses.Absolute()
        super().__init__(model=regressor, loss_function=loss_function, size=size,
            p=p, seed=seed
        )


class HardSamplingClassifier(HardSampling):
    """HardSamplingClassifier

    This wrapper enables a model to retrain on past samples who's output was hard to predict.
    This works by storing the hardest samples in a buffer of a fixed size. When a new sample
    arrives, the wrapped model is either trained on one of the buffered samples with a probability p
    or on the new sample with a probability (1 - p).

    The hardness of an observation is evaluated with a loss function that compares the sample's
    ground truth with the wrapped model's prediction. If the buffer is not full, then the sample
    is added to the buffer. If the buffer is full and the new sample has a bigger loss than the
    lowest loss in the buffer, then the sample takes it's place.

    Parameters:
        classifier (base.Classifier)
        loss_function (creme.optim.losses.CrossEntropy): Criterion used to evaluate the hardness
            of a sample.
        size (int): Size of the buffer.
        p (float): Probability of updating the model with a sample from the buffer instead of a
            new incoming sample.
        seed (int): Random seed.


    Example:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import sampling

            >>> model = preprocessing.StandardScaler()

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     sampling.HardSamplingClassifier(
            ...         classifier = linear_model.LogisticRegression(),
            ...          p = 0.1,
            ...          size = 40,
            ...         seed = 42,
            ...     )
            ... )

            >>> model_selection.progressive_val_score(
            ...     X_y = datasets.Phishing(),
            ...     model = model,
            ...     metric = metrics.ROCAUC(),
            ...     print_every = 500,
            ... )
            [500] ROCAUC: 0.928271
            [1,000] ROCAUC: 0.948547
            ROCAUC: 0.952755

    """
    def __init__(self, classifier, size, p, loss_function=None, seed=None):
        if loss_function is None:
            loss_function =  optim.losses.CrossEntropy()

        HardSampling.__init__(self, model=classifier, loss_function=loss_function, size=size,
            p=p, seed=seed
        )

    def predict_proba_one(self, x):
        return self.model.predict_proba_one(x)



"""Base interfaces.

Every estimator in ``creme`` is a class, and as such inherits from at least one base interface.
These are used to categorize, organize, and standardize the many estimators that ``creme``
contains.

"""
from .anomaly import AnomalyDetector
from .classifier import Classifier
from .classifier import BinaryClassifier
from .classifier import MultiClassifier
from .clusterer import Clusterer
from .ensemble import Ensemble
from .estimator import Estimator
from .multi_output import MultiOutputClassifier
from .multi_output import MultiOutputRegressor
from .regressor import Regressor
from .transformer import Transformer
from .wrapper import Wrapper


__all__ = [
    'AnomalyDetector',
    'BinaryClassifier',
    'Classifier',
    'Clusterer',
    'Ensemble',
    'Estimator',
    'MultiClassifier',
    'MultiOutputClassifier',
    'MultiOutputRegressor',
    'Regressor',
    'Transformer',
    'Wrapper'
]



import csv
import datetime as dt
import os
import random

from . import utils


class DictReader(csv.DictReader):
    """Overlay on top of `csv.DictReader` which allows sampling."""

    def __init__(self, fraction, rng, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fraction = fraction
        self.rng = rng

    def __next__(self):

        if self.line_num == 0:
            self.fieldnames

        row = next(self.reader)

        if self.fraction < 1:
            while self.rng.random() > self.fraction:
                row = next(self.reader)

        return dict(zip(self.fieldnames, row))


def iter_csv(filepath_or_buffer, target_name, converters=None, parse_dates=None, drop=None,
             fraction=1., compression='infer', seed=None, field_size_limit=None, **kwargs):
    """Yields rows from a CSV file.

    Parameters:
        filepath_or_buffer: Either a string indicating the location of a CSV file, or a buffer
            object that has a ``read`` method.
        target_name (str): The name of the target.
        converters (dict): A `dict` mapping feature names to callables used to parse their
            associated values.
        parse_dates (dict): A `dict` mapping feature names to a format passed to the
            `datetime.datetime.strptime` method.
        drop (list): Fields to ignore.
        fraction (float): Sampling fraction.
        compression (str): For on-the-fly decompression of on-disk data. If 'infer' and
            ``filepath_or_buffer`` is path-like, then the decompression method is inferred for the
            following extensions: '.gz', '.zip'.
        seed (int): If specified, the sampling will be deterministic.
        field_size_limit (int): If not `None`, this will passed to the `csv.field_size_limit`
            function.

    All other arguments are passed to the underlying `csv.DictReader`.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    Example:

        ::

            >>> import io
            >>> from creme import stream

            >>> data = io.StringIO('''name,day,viewers
            ... Breaking Bad,2018-03-14,1337
            ... The Sopranos,2018-03-14,42
            ... Breaking Bad,2018-03-15,7331
            ... ''')

            >>> params = dict(
            ...     target_name='viewers',
            ...     converters={'viewers': int},
            ...     parse_dates={'day': '%Y-%m-%d'}
            ... )

            >>> for x, y in stream.iter_csv(data, **params):
            ...     print(x, y)
            {'name': 'Breaking Bad', 'day': datetime.datetime(2018, 3, 14, 0, 0)} 1337
            {'name': 'The Sopranos', 'day': datetime.datetime(2018, 3, 14, 0, 0)} 42
            {'name': 'Breaking Bad', 'day': datetime.datetime(2018, 3, 15, 0, 0)} 7331

    """

    # Set the field size limit
    limit = csv.field_size_limit()
    if field_size_limit is not None:
        csv.field_size_limit(field_size_limit)

    # If a file is not opened, then we open it
    if not hasattr(filepath_or_buffer, 'read'):
        filepath_or_buffer = utils.open_filepath(filepath_or_buffer, compression)

    for x in DictReader(
        fraction=fraction,
        rng=random.Random(seed),
        f=filepath_or_buffer,
        **kwargs
    ):

        if drop:
            for i in drop:
                del x[i]

        # Cast the values to the given types
        if converters is not None:
            for i, t in converters.items():
                x[i] = t(x[i])

        # Parse the dates
        if parse_dates is not None:
            for i, fmt in parse_dates.items():
                x[i] = dt.datetime.strptime(x[i], fmt)

        # Separate the target from the features
        y = x.pop(target_name)

        yield x, y

    # Close the file
    filepath_or_buffer.close()

    # Reset the file size limit to it's original value
    csv.field_size_limit(limit)



"""Model evaluation and selection."""
import bisect
import collections
import datetime as dt
import math
import time

from .. import base
from .. import utils
from .. import stream


__all__ = ['progressive_val_score']


def progressive_val_score(X_y, model, metric, moment=None, delay=None, print_every=math.inf,
                          show_time=False, show_memory=False):
    """A variant of online scoring where the targets are revealed with a delay.

    ``X_y`` is converted into a question and answer where the model is asked to predict an
    observation. The target is only revealed to the model after a certain amount given by
    ``delay``.

    Parameters:
        X_y (generator): A stream of (features, target) tuples.
        model (base.Estimator)
        metric (metrics.Metric)
        moment (callable or str): The attribute used for measuring time. If a callable
            is passed, then it is expected to take as input a `dict` of features. If ``None``, then
            the observations are implicitely timestamped in the order in which they arrive.
        delay (callable or str or datetime.timedelta or int): The amount to wait before revealing
            the target associated with each observation to the model. This value is expected to be
            able to sum with the ``moment`` value. For instance, if ``moment`` is a
            `datetime.date`, then ``delay`` is expected to be a `datetime.timedelta`. If a callable
            is passed, then it is expected to take as input a `dict` of features and the target. If
            a `str` is passed, then it will be used to access the relevant field from the features.
            If ``None`` is passed, then no delay will be used, which leads to doing standard online
            validation.
        print_every (int): Iteration number at which to print the current metric. This only takes
            into account the predictions, and not the training steps.
        show_time (bool): Whether or not to display the elapsed time.
        show_memory (bool): Whether or not to display the memory usage of the model.

    Returns:
        metrics.Metric

    References:
        1. `Beating the Hold-Out: Bounds for K-fold and Progressive Cross-Validation <http://hunch.net/~jl/projects/prediction_bounds/progressive_validation/coltfinal.pdf>`_
        2. `Grzenda, M., Gomes, H.M. and Bifet, A., 2019. Delayed labelling evaluation for data streams. Data Mining and Knowledge Discovery, pp.1-30. <https://link.springer.com/content/pdf/10.1007%2Fs10618-019-00654-y.pdf>`_

    """

    # Check that the model and the metric are in accordance
    if not metric.works_with(model):
        raise ValueError(f'{metric.__class__.__name__} metric is not compatible with {model}')

    # Determine if predict_one or predict_proba_one should be used in case of a classifier
    pred_func = model.predict_one
    is_classifier = isinstance(utils.estimator_checks.guess_model(model), base.Classifier)
    if is_classifier and not metric.requires_labels:
        pred_func = model.predict_proba_one

    preds = {}

    n_total_answers = 0
    if show_time:
        start = time.perf_counter()

    for i, x, y in stream.simulate_qa(X_y, moment, delay, copy=True):

        # Question
        if y is None:
            preds[i] = pred_func(x=x)

        # Answer
        else:
            y_pred = preds.pop(i)
            if y_pred != {} and y_pred is not None:
                metric.update(y_true=y, y_pred=y_pred)
            model.fit_one(x=x, y=y)

            # Update the answer counter
            n_total_answers += 1
            if not n_total_answers % print_every:
                msg = f'[{n_total_answers:,d}] {metric}'
                if show_time:
                    now = time.perf_counter()
                    msg += f' – {dt.timedelta(seconds=int(now - start))}'
                if show_memory:
                    msg += f' – {model._memory_usage}'
                print(msg)

    return metric



from .. import base


__all__ = ['Discard', 'Select']


class Discard(base.Transformer):
    """Subsets a set of features by applying a blacklist.

    Parameters:
        blacklist (strs): Key(s) to discard.

    Example:

        ::

            >>> from creme import compose

            >>> x = {'a': 42, 'b': 12, 'c': 13}
            >>> compose.Discard('a', 'b').transform_one(x)
            {'c': 13}

    """

    def __init__(self, *blacklist):
        self.blacklist = set(blacklist)

    def transform_one(self, x):
        return {i: xi for i, xi in x.items() if i not in self.blacklist}

    def __str__(self):
        return '~' + str(sorted(self.blacklist))

    def __repr__(self):
        if self.blacklist:
            return 'Discard (\n  ' + '\n  '.join(map(str, sorted(self.blacklist))) + '\n)'
        return 'Discard ()'

    def _set_params(self, blacklist=None):
        if not blacklist:
            blacklist = self.blacklist
        return Select(*blacklist)


class Select(base.Transformer):
    """Subsets a set of features by applying a whitelist.

    Parameters:
        whitelist (strs): Key(s) to keep.

    Example:

        ::

            >>> from creme import compose

            >>> x = {'a': 42, 'b': 12, 'c': 13}
            >>> compose.Select('c').transform_one(x)
            {'c': 13}

    """

    def __init__(self, *whitelist):
        self.whitelist = set(whitelist)

    def transform_one(self, x):
        return {i: x[i] for i in self.whitelist}

    def __str__(self):
        return str(sorted(self.whitelist))

    def __repr__(self):
        if self.whitelist:
            return 'Select (\n  ' + '\n  '.join(map(str, sorted(self.whitelist))) + '\n)'
        return 'Select ()'

    def _get_params(self):
        return self.whitelist

    def _set_params(self, whitelist=None):
        if not whitelist:
            whitelist = self.whitelist
        return Select(*whitelist)



import abc

from . import estimator


class Clusterer(estimator.Estimator):
    """A clusterer."""

    @abc.abstractmethod
    def fit_one(self, x: dict) -> 'Clusterer':
        """Fits to a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> int:
        """Predicts the cluster number of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            int

        """



collect_ignore = []

try:
    import sklearn
except ImportError:
    collect_ignore.append('compat/sklearn.py')
    collect_ignore.append('compat/test_sklearn.py')

try:
    import surprise
except ImportError:
    collect_ignore.append('reco/surprise.py')


try:
    import torch
except ImportError:
    collect_ignore.append('compat/pytorch.py')



import abc

from .. import base


class Recommender(base.Regressor):
    """A recommender."""

    def fit_one(self, x, y):
        return self._fit_one(x['user'], x['item'], y)

    def predict_one(self, x):
        return self._predict_one(x['user'], x['item'])

    @abc.abstractmethod
    def _fit_one(self, user, item, y):
        """Fits a ``user``-``item`` pair and a real-valued target ``y``.

        Parameters:
            user (int or str)
            item (int or str)
            y (float)

        Returns:
            self: object

        """

    @abc.abstractmethod
    def _predict_one(self, user, item):
        """Predicts the target value of a set of features ``x``.

        Parameters:
            user (int or str)
            item (int or str)

        Returns:
            float: The prediction.

        """



import abc
import typing

from . import estimator


class Classifier(estimator.Estimator):
    """A classifier."""

    @abc.abstractmethod
    def fit_one(self, x, y):
        """Fits to a set of features ``x`` and a label ``y``.

        Parameters:
            x (dict)
            y (Label)

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_proba_one(self, x):
        """Predicts the probability output of a set of features ``x``."""

    def predict_one(self, x: dict) -> typing.Union[typing.Hashable, None]:
        """Predicts the target value of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            Label

        """
        y_pred = self.predict_proba_one(x)
        if y_pred:
            return max(y_pred, key=y_pred.get)
        return None


class MultiClassifier(Classifier):
    """A multi-class classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: typing.Hashable) -> 'Classifier':
        """Fits to a set of features ``x`` and a label ``y``.

        Parameters:
            x (dict)
            y (Label)

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_proba_one(self, x: dict) -> typing.Dict[typing.Hashable, float]:
        """Predicts the probability output of a set of features ``x``."""


class BinaryClassifier(Classifier):
    """A binary classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: bool) -> 'BinaryClassifier':
        """Fits to a set of features ``x`` and a boolean target ``y``.

        Parameters:
            x (dict)
            y (bool)

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_proba_one(self, x: dict) -> typing.Dict[bool, float]:
        """Predicts the probability output of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            dict of floats

        """

    def predict_one(self, x: dict) -> bool:
        """Predicts the target value of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            bool

        """
        y_pred = self.predict_proba_one(x)
        return y_pred[True] > y_pred[False]



from ... import ensemble

from . import tree


__all__ = ['RandomForestClassifier']


class RandomForestClassifier(ensemble.BaggingClassifier):
    """Random forest classifier.

    This is a thin wrapper over `ensemble.BaggingClassifier` and `tree.DecisionTreeClassifier`.

    Parameters:
        n_trees (int): The number of trees in the forest.
        seed (int): Random number generator seed for reproducibility.
        **tree_params (dict): The parameters of the decision tree.

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import tree

            >>> X_y = datasets.Phishing()

            >>> model = tree.RandomForestClassifier(
            ...     n_trees=10,
            ...     seed=42,
            ...     # Tree parameters
            ...     patience=100,
            ...     confidence=1e-5,
            ...     criterion='gini'
            ... )

            >>> metric = metrics.LogLoss()

            >>> model_selection.progressive_val_score(X_y, model, metric)
            LogLoss: 0.456533

    """

    def __init__(self, n_trees=10, seed=None, **tree_params):
        super().__init__(
            model=tree.DecisionTreeClassifier(**tree_params),
            n_models=n_trees,
            seed=seed
        )

    @property
    def n_trees(self):
        return len(self)

    def _get_params(self):
        return {**super()._get_params(), **self.model._get_params()}

    def __str__(self):
        return 'RandomForestClassifier'



from .. import stream

from . import base


class Taxis(base.FileDataset):
    """Taxi ride durations in New York City.

    The goal is to predict the duration of taxi rides in New York City.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `New York City Taxi Trip Duration competition on Kaggle <https://www.kaggle.com/c/nyc-taxi-trip-duration>`_

    """

    def __init__(self, data_home=None, verbose=True):
        super().__init__(
            n_samples=1_458_644,
            n_features=8,
            category=base.REG,
            url='https://maxhalford.github.io/files/datasets/nyc_taxis.zip',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            f'{directory}/train.csv',
            target_name='trip_duration',
            converters={
                'passenger_count': int,
                'pickup_longitude': float,
                'pickup_latitude': float,
                'dropoff_longitude': float,
                'dropoff_latitude': float,
                'trip_duration': int
            },
            parse_dates={'pickup_datetime': '%Y-%m-%d %H:%M:%S'},
            drop=['dropoff_datetime', 'id']
        )



import collections

from .. import utils

from . import base
from . import accuracy
from . import fbeta
from . import precision
from . import recall


class ClassificationReport(base.MultiClassMetric):
    """A report for monitoring a classifier.

    This class maintains a set of metrics and updates each of them every time `update` is called.
    You can print this class at any time during a model's lifetime to get a tabular visualization
    of various metrics.

    Parameters:
        digits (int): The number of decimals to display for each metric.

    Attributes:
        f1s (collections.defaultdict): Contains an instance of `metrics.F1` for each label.
        macro_precision (metrics.MacroPrecision)
        macro_recall (metrics.MacroRecall)
        macro_f1 (metrics.MacroF1)
        micro_precision (metrics.MicroPrecision)
        micro_recall (metrics.MicroRecall)
        micro_f1 (metrics.MicroF1)
        weighted_precision (metrics.WeightedPrecision)
        weighted_recall (metrics.WeightedRecall)
        weighted_f1 (metrics.WeightedF1)
        accuracy (metrics.Accuracy)
        support (collections.Counter): Records the number of times a label is encountered.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = ['pear', 'apple', 'banana', 'banana', 'banana']
            >>> y_pred = ['apple', 'pear', 'banana', 'banana', 'apple']

            >>> report = metrics.ClassificationReport()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     report = report.update(yt, yp)

            >>> print(report)
            ... # doctest: +NORMALIZE_WHITESPACE
                       Precision   Recall   F1      Support
            <BLANKLINE>
               apple       0.000    0.000   0.000         1
              banana       1.000    0.667   0.800         3
                pear       0.000    0.000   0.000         1
            <BLANKLINE>
               Macro       0.333    0.222   0.267
               Micro       0.400    0.400   0.400
            Weighted       0.600    0.400   0.480
            <BLANKLINE>
                             40.0% accuracy

    Note:
        You can wrap a `metrics.ClassificationReport` with `metrics.Rolling` in order to obtain
        a classification report over a recent window of observations.

    """

    def __init__(self, decimals=3):
        self.decimals = decimals
        self.f1s = collections.defaultdict(fbeta.F1)
        self.macro_precision = precision.MacroPrecision()
        self.macro_recall = recall.MacroRecall()
        self.macro_f1 = fbeta.MacroF1()
        self.micro_precision = precision.MicroPrecision()
        self.micro_recall = recall.MicroRecall()
        self.micro_f1 = fbeta.MicroF1()
        self.weighted_precision = precision.WeightedPrecision()
        self.weighted_recall = recall.WeightedRecall()
        self.weighted_f1 = fbeta.WeightedF1()
        self.accuracy = accuracy.Accuracy()
        self.support = collections.Counter()

    @property
    def bigger_is_better(self):
        return True

    @property
    def get(self):
        raise NotImplementedError

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred, sample_weight=1.):

        self.support[y_true] += 1

        # Update per class metrics
        for c in self.support:
            self.f1s[c].update(y_true == c, y_pred == c, sample_weight)

        # Update global metrics
        for m in [self.macro_precision, self.macro_recall, self.macro_f1,
                  self.micro_precision, self.micro_recall, self.micro_f1,
                  self.weighted_precision, self.weighted_recall, self.weighted_f1]:
            m.update(y_true, y_pred, sample_weight)

        self.accuracy.update(y_true, y_pred)

        return self

    def revert(self, y_true, y_pred, sample_weight=1.):

        self.support[y_true] -= 1

        # Revert per class metrics
        for c in self.support:
            self.f1s[c].revert(y_true == c, y_pred == c, sample_weight)

        # Revert global metrics
        for m in [self.macro_precision, self.macro_recall, self.macro_f1,
                  self.micro_precision, self.micro_recall, self.micro_f1,
                  self.weighted_precision, self.weighted_recall, self.weighted_f1]:
            m.revert(y_true, y_pred, sample_weight)
        self.accuracy.revert(y_true, y_pred)

        return self

    def __repr__(self):

        def fmt_float(x):
            return f'{x:.{self.decimals}f}'

        headers = ['', 'Precision', 'Recall', 'F1', 'Support']
        classes = sorted(self.support.keys())
        columns = [
            # Row names
            ['', *map(str, classes), '', 'Macro', 'Micro', 'Weighted'],
            # Precision values
            [
                '', *[fmt_float(self.f1s[c].precision.get()) for c in classes], '',
                *map(fmt_float, [
                    self.macro_precision.get(),
                    self.micro_precision.get(),
                    self.weighted_precision.get()
                ])
            ],
            # Recall values
            [
                '', *[fmt_float(self.f1s[c].recall.get()) for c in classes], '',
                *map(fmt_float, [
                    self.macro_recall.get(),
                    self.micro_recall.get(),
                    self.weighted_recall.get()
                ])
            ],
            # F1 values
            [
                '', *[fmt_float(self.f1s[c].get()) for c in classes], '',
                *map(fmt_float, [
                    self.macro_f1.get(),
                    self.micro_f1.get(),
                    self.weighted_f1.get()
                ])
            ],
            # Support
            ['', *[str(self.support[c]) for c in classes], *[''] * 4]
        ]

        # Build the table
        table = utils.pretty.print_table(headers, columns)

        # Write down the accuracy
        width = len(table.splitlines()[0])
        accuracy = f'{self.accuracy.get():.{self.decimals - 2}%}' + ' accuracy'
        table += '\n\n' + f'{accuracy:^{width}}'

        return table



import itertools
import types

import numpy as np


__all__ = ['expand_param_grid']


def expand_param_grid(grid):
    """Expands a grid of possible parameters into a list of single parametrizations.

    Parameter:
        grid (dict or list of dicts)

    Example:

        >>> import pprint
        >>> from creme import model_selection
        >>> from creme import optim

        >>> param_grid = {
        ...     'BagOfWords': {
        ...         'strip_accents': [False, True]
        ...     },
        ...     'LinearRegression': {
        ...         'intercept_lr': [0.001],
        ...         'optimizer': (optim.SGD, {'lr': [0.001, 0.01, 0.1]}),
        ...     }
        ... }

        >>> for params in model_selection.expand_param_grid(param_grid):
        ...     pprint.pprint(params)
        ...     print()
        {'BagOfWords': {'strip_accents': False},
         'LinearRegression': {'intercept_lr': 0.001,
                              'optimizer': SGD({'lr': Constant({'learning_rate': 0.001}), 'n_iterations': 0})}}
        <BLANKLINE>
        {'BagOfWords': {'strip_accents': False},
         'LinearRegression': {'intercept_lr': 0.001,
                              'optimizer': SGD({'lr': Constant({'learning_rate': 0.01}), 'n_iterations': 0})}}
        <BLANKLINE>
        {'BagOfWords': {'strip_accents': False},
         'LinearRegression': {'intercept_lr': 0.001,
                              'optimizer': SGD({'lr': Constant({'learning_rate': 0.1}), 'n_iterations': 0})}}
        <BLANKLINE>
        {'BagOfWords': {'strip_accents': True},
         'LinearRegression': {'intercept_lr': 0.001,
                              'optimizer': SGD({'lr': Constant({'learning_rate': 0.001}), 'n_iterations': 0})}}
        <BLANKLINE>
        {'BagOfWords': {'strip_accents': True},
         'LinearRegression': {'intercept_lr': 0.001,
                              'optimizer': SGD({'lr': Constant({'learning_rate': 0.01}), 'n_iterations': 0})}}
        <BLANKLINE>
        {'BagOfWords': {'strip_accents': True},
         'LinearRegression': {'intercept_lr': 0.001,
                              'optimizer': SGD({'lr': Constant({'learning_rate': 0.1}), 'n_iterations': 0})}}
        <BLANKLINE>

    """

    def expand_tuple(t):

        klass, params = t

        if not isinstance(klass, types.ClassType):
            raise ValueError(f'Expected first element to be a class, got {klass}')

        if not isinstance(params, dict):
            raise ValueError(f'Expected second element to be a dict, got {params}')

        return (klass(**combo) for combo in expand_param_grid(params))

    def expand(k, v):

        if isinstance(v, tuple):
            return ((k, el) for el in expand_tuple(v))

        # Example:
        # k = 'lr'
        # v = [0.001, 0.01, 0.1]
        if isinstance(v, (list, set, np.ndarray)):

            combos = []

            for el in v:
                if isinstance(el, tuple):
                    for combo in expand_tuple(el):
                        combos.append((k, combo))
                else:
                    combos.append((k, el))

            return combos

        if isinstance(v, dict):

            # Example:
            # k = 'LinearRegression'
            # v = {
            #     'intercept_lr': [0.001],
            #     'l2': [1],
            #     'optimizer': {
            #         optim.Adam: {
            #             'beta_1': [0.1, 0.01, 0.001],
            #             'lr': [0.1, 0.01, 0.001]
            #         },
            #      }
            # }
            return ((k, el) for el in expand_param_grid(v))

        raise ValueError(f'unsupported type: {type(v)}')

    for key in grid:
        if not isinstance(key, str):
            raise ValueError(f'Expected a key of type str; got {key}')

    # Example:
    # grid = {
    #     'beta_1': [0.1, 0.01, 0.001],
    #     'lr': [0.1, 0.01, 0.001]
    # }
    return (
        dict(el) if isinstance(el[0], tuple) else el[0]
        for el in itertools.product(*(expand(k, v) for k, v in grid.items()))
    )



import itertools
import random
import types


def shuffle(stream, buffer_size, seed=None):
    """Shuffles a stream of data.

    This works by maintaining a buffer of elements. The first buffer_size elements are stored in
    memory. Once the buffer is full, a random element inside the buffer is yielded. Everytime an
    element is yielded, the next element in the stream replaces it and the buffer is sampled again.
    Increasing buffer_size will improve the quality of the shuffling.

    If you really want to stream over your dataset in a "good" random order, the best way is to
    split your dataset into smaller datasets and loop over them in a round-robin fashion. You may
    do this by using the ``roundrobin`` recipe from the `itertools` module.

    Example:

        ::

            >>> from creme import stream

            >>> for i in stream.shuffle(range(15), buffer_size=5, seed=42):
            ...     print(i)
            0
            5
            2
            1
            8
            9
            6
            4
            11
            12
            10
            7
            14
            13
            3

    References:
        1. `Visualizing TensorFlow's streaming shufflers <http://www.moderndescartes.com/essays/shuffle_viz/>`_

    """

    rng = random.Random(seed)

    # If stream is not a generator, then we coerce it to one
    if not isinstance(stream, types.GeneratorType):
        stream = iter(stream)

    # Initialize the buffer with the first buffer_size elements of the stream
    buffer = list(itertools.islice(stream, buffer_size))

    # Deplete the stream until it is empty
    for element in stream:

        # Pick a random element from the buffer and yield it
        i = rng.randint(0, len(buffer) - 1)
        yield buffer[i]

        # Replace the yielded element from the buffer with the new element from the stream
        buffer[i] = element

    # Shuffle the remaining buffer elements and yield them one by one
    rng.shuffle(buffer)
    yield from buffer



import copy
import functools
import math
import operator

from .. import base


__all__ = ['SuccessiveHalvingClassifier', 'SuccessiveHalvingRegressor']


class SuccessiveHalving:

    def __init__(self, models, metric, budget, eta=2, verbose=False):

        # Check that the model and the metric are in accordance
        for model in models:
            if not metric.works_with(model):
                raise ValueError(f"{metric.__class__.__name__} metric can't be used to evaluate a " +
                                f'{model.__class__.__name__}')

        self.models = models
        self.metric = metric
        self.budget = budget
        self.eta = eta
        self.verbose = verbose

        self._n = len(models)
        self._metrics = [copy.deepcopy(metric) for _ in range(self._n)]
        self._s = self._n
        self._n_rungs = 0
        self._rankings = list(range(self._n))
        self._r = math.floor(budget / (self._s * math.ceil(math.log(self._n, eta))))
        self._budget_used = 0
        self._n_iterations = 0
        self._best_model_idx = 0

        if isinstance(model, base.Classifier) and not metric.requires_labels:
            self._pred_func = lambda model: model.predict_proba_one
        else:
            self._pred_func = lambda model: model.predict_one

    @property
    def best_model(self):
        """The current best model."""
        return self.models[self._best_model_idx]

    def fit_one(self, x, y):

        for i in self._rankings[:self._s]:
            model = self.models[i]
            metric = self._metrics[i]
            y_pred = self._pred_func(model)(x)
            metric.update(y_true=y, y_pred=y_pred)
            model.fit_one(x, y)

            # Check for a new best model
            if i != self._best_model_idx:
                op = operator.gt if self.metric.bigger_is_better else operator.lt
                if op(self._metrics[i].get(), self._metrics[self._best_model_idx].get()):
                    self._best_model_idx = i

        self._n_iterations += 1

        if self._s > 1 and self._n_iterations == self._r:

            self._n_rungs += 1
            self._budget_used += self._s * self._r

            # Update the rankings of the current models based on their respective metric values
            self._rankings[:self._s] = sorted(
                self._rankings[:self._s],
                key=lambda i: self._metrics[i].get(),
                reverse=self.metric.bigger_is_better
            )

            # Determine how many models to keep for the current rung
            cutoff = math.ceil(self._s / self.eta)

            if self.verbose:
                print('\t'.join((
                    f'[{self._n_rungs}]',
                    f'{self._s - cutoff} removed',
                    f'{cutoff} left',
                    f'{self._r} iterations',
                    f'budget used: {self._budget_used}',
                    f'budget left: {self.budget - self._budget_used}',
                    f'best {self._metrics[self._rankings[0]]}',
                )))

            # Determine where the next rung is located
            self._s = cutoff
            self._r = math.floor(self.budget / (self._s * math.ceil(math.log(self._n, self.eta))))

        return self


class SuccessiveHalvingRegressor(SuccessiveHalving, base.Regressor):
    """Successive halving algorithm for regression.

    Successive halving is a method for performing model selection without having to train each
    model on all the dataset. At certain points in time (called "rungs"), the worst performing will
    be discarded and the best ones will keep competing between each other. The rung values are
    designed so that at most ``budget`` model updates will be performed in total.

    If you have ``k`` combinations of hyperparameters and that your dataset contains ``n``
    observations, then the maximal budget you can allocate is:

    .. math:: \\frac{2kn}{eta}

    It is recommended that you check this beforehand. This bound can't be checked by the function
    because the size of the dataset is not known. In fact it is potentially infinite, in which case
    the algorithm will terminate once all the budget has been spent.

    If you have a budget of ``B``, and that your dataset contains ``n`` observations, then the
    number of hyperparameter combinations that will spend all the budget and go through all the
    data is:

    .. math:: \\ceil(\\floor(\\frac{B}{(2n)}) \times eta)

    Parameters:
        model (models.Regressor)
        param_grid (dict): Parameter grid.
        X_y (generator): A stream of (features, target) tuples.
        metric (metrics.RegressionMetric): Metric used for comparing models with.
        budget (int): Total number of model updates you wish to allocate.
        eta (int): Rate of elimination. At every rung, ``math.ceil(k / eta)`` models are kept,
            where ``k`` is the number of models that have reached the rung. A higher ``eta`` value
            will focus on less models but will allocate more iterations to the best models.
        verbose (bool): Whether to display progress or not.

    Examples:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import preprocessing
            >>> from creme import model_selection
            >>> from creme import optim

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LinearRegression(intercept_lr=.1)
            ... )

            >>> param_grid = model_selection.expand_param_grid({
            ...     'LinearRegression': {
            ...         'optimizer': [
            ...             (optim.SGD, {'lr': [0.1, 0.01, 0.005]}),
            ...             (optim.Adam, {'beta_1': [0.01, 0.001], 'lr': [0.1, 0.01, 0.001]}),
            ...             (optim.Adam, {'beta_1': [0.1], 'lr': [0.001]}),
            ...         ]
            ...     }
            ... })

            >>> models = [model._set_params(params) for params in param_grid]

            >>> sh = model_selection.SuccessiveHalvingRegressor(
            ...     models=models,
            ...     metric=metrics.MAE(),
            ...     budget=2000,
            ...     eta=2,
            ...     verbose=True
            ... )

            >>> model_selection.progressive_val_score(
            ...     X_y=datasets.TrumpApproval(),
            ...     model=sh,
            ...     metric=metrics.MAE()
            ... )
            [1]	5 removed	5 left	50 iterations	budget used: 500	budget left: 1500	best MAE: 4.541564
            [2]	2 removed	3 left	100 iterations	budget used: 1000	budget left: 1000	best MAE: 2.453893
            [3]	1 removed	2 left	166 iterations	budget used: 1498	budget left: 502	best MAE: 1.579851
            [4]	1 removed	1 left	250 iterations	budget used: 1998	budget left: 2	    best MAE: 1.146581
            MAE: 0.472907

            >>> sh.best_model
            Pipeline (
              StandardScaler (
                with_mean=True
                with_std=True
              ),
              LinearRegression (
                optimizer=Adam (
                  lr=Constant (
                    learning_rate=0.1
                  )
                  beta_1=0.001
                  beta_2=0.999
                  eps=1e-08
                )
                loss=Squared ()
                l2=0.
                intercept=39.838695
                intercept_lr=Constant (
                  learning_rate=0.1
                )
                clip_gradient=1e+12
                initializer=Zeros ()
              )
            )

    References:
        1. `Jamieson, K. and Talwalkar, A., 2016, May. Non-stochastic best arm identification and hyperparameter optimization. In Artificial Intelligence and Statistics (pp. 240-248). <http://proceedings.mlr.press/v51/jamieson16.pdf>`_
        2. `Li, L., Jamieson, K., Rostamizadeh, A., Gonina, E., Hardt, M., Recht, B. and Talwalkar, A., 2018. Massively parallel hyperparameter tuning. arXiv preprint arXiv:1810.05934. <https://arxiv.org/pdf/1810.05934.pdf>`_
        3. `Li, L., Jamieson, K., DeSalvo, G., Rostamizadeh, A. and Talwalkar, A., 2017. Hyperband: A novel bandit-based approach to hyperparameter optimization. The Journal of Machine Learning Research, 18(1), pp.6765-6816. <https://arxiv.org/pdf/1603.06560.pdf>`_

    """

    def predict_one(self, x):
        return self.best_model.predict_one(x)


class SuccessiveHalvingClassifier(SuccessiveHalving, base.Classifier):
    """Successive halving algorithm for classification.

    Successive halving is a method for performing model selection without having to train each
    model on all the dataset. At certain points in time (called "rungs"), the worst performing will
    be discarded and the best ones will keep competing between each other. The rung values are
    designed so that at most ``budget`` model updates will be performed in total.

    If you have ``k`` combinations of hyperparameters and that your dataset contains ``n``
    observations, then the maximal budget you can allocate is:

    .. math:: \\frac{2kn}{eta}

    It is recommended that you check this beforehand. This bound can't be checked by the function
    because the size of the dataset is not known. In fact it is potentially infinite, in which case
    the algorithm will terminate once all the budget has been spent.

    If you have a budget of ``B``, and that your dataset contains ``n`` observations, then the
    number of hyperparameter combinations that will spend all the budget and go through all the
    data is:

    .. math:: \\ceil(\\floor(\\frac{B}{(2n)}) \times eta)

    Parameters:
        model (models.Classifier)
        param_grid (dict): Parameter grid.
        X_y (generator): A stream of (features, target) tuples.
        metric (metrics.ClassificationMetric): Metric used for comparing models with.
        budget (int): Total number of model updates you wish to allocate.
        eta (int): Rate of elimination. At every rung, ``math.ceil(k / eta)`` models are kept,
            where ``k`` is the number of models that have reached the rung. A higher ``eta`` value
            will focus on less models but will allocate more iterations to the best models.
        verbose (bool): Whether to display progress or not.

    Examples:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression()
            ... )

            >>> param_grid = model_selection.expand_param_grid({
            ...     'LogisticRegression': {
            ...         'optimizer': [
            ...             (optim.SGD, {'lr': [0.1, 0.01, 0.005]}),
            ...             (optim.Adam, {'beta_1': [0.01, 0.001], 'lr': [0.1, 0.01, 0.001]}),
            ...             (optim.Adam, {'beta_1': [0.1], 'lr': [0.001]}),
            ...         ]
            ...     }
            ... })

            >>> models = [model._set_params(params) for params in param_grid]

            >>> sh = model_selection.SuccessiveHalvingClassifier(
            ...     models=models,
            ...     metric=metrics.Accuracy(),
            ...     budget=2000,
            ...     eta=2,
            ...     verbose=True
            ... )

            >>> model_selection.progressive_val_score(
            ...     X_y=datasets.Phishing(),
            ...     model=sh,
            ...     metric=metrics.ROCAUC()
            ... )
            [1]	5 removed	5 left	50 iterations	budget used: 500	budget left: 1500	best Accuracy: 80.00%
            [2]	2 removed	3 left	100 iterations	budget used: 1000	budget left: 1000	best Accuracy: 84.00%
            [3]	1 removed	2 left	166 iterations	budget used: 1498	budget left: 502	best Accuracy: 86.75%
            [4]	1 removed	1 left	250 iterations	budget used: 1998	budget left: 2	    best Accuracy: 85.20%
            ROCAUC: 0.949693

            >>> sh.best_model
            Pipeline (
              StandardScaler (
                with_mean=True
                with_std=True
              ),
              LogisticRegression (
                optimizer=Adam (
                  lr=Constant (
                    learning_rate=0.01
                  )
                  beta_1=0.01
                  beta_2=0.999
                  eps=1e-08
                )
                loss=Log ()
                l2=0.
                intercept=-0.394786
                intercept_lr=Constant (
                  learning_rate=0.01
                )
                clip_gradient=1e+12
                initializer=Zeros ()
              )
            )

    References:
        1. `Jamieson, K. and Talwalkar, A., 2016, May. Non-stochastic best arm identification and hyperparameter optimization. In Artificial Intelligence and Statistics (pp. 240-248). <http://proceedings.mlr.press/v51/jamieson16.pdf>`_
        2. `Li, L., Jamieson, K., Rostamizadeh, A., Gonina, E., Hardt, M., Recht, B. and Talwalkar, A., 2018. Massively parallel hyperparameter tuning. arXiv preprint arXiv:1810.05934. <https://arxiv.org/pdf/1810.05934.pdf>`_
        3. `Li, L., Jamieson, K., DeSalvo, G., Rostamizadeh, A. and Talwalkar, A., 2017. Hyperband: A novel bandit-based approach to hyperparameter optimization. The Journal of Machine Learning Research, 18(1), pp.6765-6816. <https://arxiv.org/pdf/1603.06560.pdf>`_

    """

    def predict_proba_one(self, x):
        return self.best_model.predict_proba_one(x)



from .. import stream

from . import base


class Elec2(base.FileDataset):
    """Electricity prices in New South Wales.

    This data was collected from the Australian New South Wales Electricity Market. In this market,
    prices are not fixed and are affected by demand and supply of the market. They are set every
    five minutes. Electricity transfers to/from the neighboring state of Victoria were done to
    alleviate fluctuations.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `SPLICE-2 Comparative Evaluation: Electricity Pricing <http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.12.9405>`_
        2. `DataHub description <https://datahub.io/machine-learning/electricity#readme>`_

    """

    def __init__(self, data_home=None, verbose=False):
        super().__init__(
            n_samples=45_312,
            n_features=8,
            category=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/electricity.zip',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            f'{directory}/electricity.csv',
            target_name='class',
            converters={
                'date': float,
                'day': int,
                'period': float,
                'nswprice': float,
                'nswdemand': float,
                'vicprice': float,
                'vicdemand': float,
                'transfer': float,
                'class': lambda x: x == 'UP'
            }
        )



import collections
import math
import numbers

import numpy as np

from .. import base
from .. import optim
from .. import utils


__all__ = [
    'LinearRegression',
    'LogisticRegression'
]


class GLM:
    """Generalized Linear Model.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately.
        loss (optim.Loss): The loss function to optimize for.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        intercept (float): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        clip_gradient (float): Clips the absolute value of each gradient value.
        initializer (optim.initializers.Initializer): Weights initialization scheme.

    Attributes:
        weights (collections.defaultdict): The current weights.

    """

    def __init__(self, optimizer, loss, l2, intercept, intercept_lr, clip_gradient, initializer):
        self.optimizer = optimizer
        self.loss = loss
        self.l2 = l2
        self.intercept = intercept
        self.intercept_lr = (
            optim.schedulers.Constant(intercept_lr)
            if isinstance(intercept_lr, numbers.Number) else
            intercept_lr
        )
        self.clip_gradient = clip_gradient
        self.weights = collections.defaultdict(initializer)
        self.initializer = initializer

    def _raw_dot(self, x):
        return utils.math.dot(self.weights, x) + self.intercept

    def _eval_gradient(self, x, y, sample_weight):
        """Returns the gradient for a given observation.

        This logic is put into a separate function for testing purposes.

        """

        loss_gradient = self.loss.gradient(y_true=y, y_pred=self._raw_dot(x))

        # Apply the sample weight
        loss_gradient *= sample_weight

        # Clip the gradient to avoid numerical instability
        loss_gradient = utils.math.clamp(
            loss_gradient,
            minimum=-self.clip_gradient,
            maximum=self.clip_gradient
        )

        return (
            {
                i: (
                    xi * loss_gradient +
                    2. * self.l2 * self.weights.get(i, 0)
                )
                for i, xi in x.items()
            },
            loss_gradient
        )

    def fit_one(self, x, y, sample_weight=1.):

        # Some optimizers need to do something before a prediction is made
        self.weights = self.optimizer.update_before_pred(w=self.weights)

        # Calculate the gradient
        gradient, loss_gradient = self._eval_gradient(x=x, y=y, sample_weight=sample_weight)

        # Update the intercept
        self.intercept -= self.intercept_lr.get(self.optimizer.n_iterations) * loss_gradient

        # Update the weights
        self.weights = self.optimizer.update_after_pred(w=self.weights, g=gradient)

        return self


class LinearRegression(GLM, base.Regressor):
    """Linear regression.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately. Defaults to ``optim.SGD(.01)``.
        loss (optim.RegressionLoss): The loss function to optimize for. Defaults to
            ``optim.losses.SquaredLoss``.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        intercept (float): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        clip_gradient (float): Clips the absolute value of each gradient value.
        initializer (optim.initializers.Initializer): Weights initialization scheme.

    Attributes:
        weights (collections.defaultdict): The current weights.

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing

            >>> X_y = datasets.TrumpApproval()

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LinearRegression(intercept_lr=.1)
            ... )
            >>> metric = metrics.MAE()

            >>> model_selection.progressive_val_score(X_y, model, metric)
            MAE: 0.616405

            >>> model['LinearRegression'].intercept
            38.000439

    Note:
        Using a feature scaler such as `preprocessing.StandardScaler` upstream helps the optimizer
        to converge.

    """

    def __init__(self, optimizer=None, loss=None, l2=.0, intercept=0., intercept_lr=.01,
                 clip_gradient=1e12, initializer=None):
        super().__init__(
            optimizer=(
                optim.SGD(optim.schedulers.InverseScaling(.01, .25))
                if optimizer is None else
                optimizer
            ),
            loss=optim.losses.Squared() if loss is None else loss,
            intercept=intercept,
            intercept_lr=intercept_lr,
            l2=l2,
            clip_gradient=clip_gradient,
            initializer=initializer if initializer else optim.initializers.Zeros()
        )

    def predict_one(self, x):
        return self.loss.mean_func(self._raw_dot(x))

    def debug_one(self, x, decimals=5, **print_params):
        """

        Example:

            ::

                >>> from creme import datasets
                >>> from creme import linear_model
                >>> from creme import metrics
                >>> from creme import model_selection
                >>> from creme import preprocessing

                >>> X_y = datasets.TrumpApproval()

                >>> model = (
                ...     preprocessing.StandardScaler() |
                ...     linear_model.LinearRegression(intercept_lr=.1)
                ... )

                >>> for x, y in X_y:
                ...     y_pred = model.predict_one(x)
                ...     model = model.fit_one(x, y)

                >>> model.debug_one(x)
                0. Input
                --------
                gallup: 43.84321 (float)
                ipsos: 40.57068 (float)
                morning_consult: 37.81875 (float)
                ordinal_date: 737389 (int)
                rasmussen: 40.10469 (float)
                you_gov: 41.63691 (float)
                <BLANKLINE>
                1. StandardScaler
                -----------------
                gallup: 1.18751 (float)
                ipsos: -0.04683 (float)
                morning_consult: -1.22583 (float)
                ordinal_date: 1.72946 (float)
                rasmussen: -0.23857 (float)
                you_gov: 0.44131 (float)
                <BLANKLINE>
                2. LinearRegression
                -------------------
                Name              Value      Weight      Contribution
                      Intercept    1.00000    38.00044       38.00044
                   ordinal_date    1.72946     2.23125        3.85885
                         gallup    1.18751     0.28647        0.34019
                        you_gov    0.44131    -0.01270       -0.00560
                          ipsos   -0.04683     1.01815       -0.04768
                      rasmussen   -0.23857     0.45099       -0.10759
                morning_consult   -1.22583     0.35181       -0.43126
                <BLANKLINE>
                Prediction: 41.60735

        """

        def fmt_float(x):
            return '{: ,.{prec}f}'.format(x, prec=decimals)

        names = list(map(str, x.keys())) + ['Intercept']
        values = list(map(fmt_float, list(x.values()) + [1]))
        weights = list(map(fmt_float, [self.weights.get(i, 0) for i in x] + [self.intercept]))
        contributions = [xi * self.weights.get(i, 0) for i, xi in x.items()] + [self.intercept]
        order = reversed(np.argsort(contributions))
        contributions = list(map(fmt_float, contributions))

        table = utils.pretty.print_table(
            headers=['Name', 'Value', 'Weight', 'Contribution'],
            columns=[names, values, weights, contributions],
            order=order
        )

        print(table, **print_params)


class LogisticRegression(GLM, base.BinaryClassifier):
    """Logistic regression.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately. Defaults to ``optim.SGD(.05)``.
        loss (optim.BinaryLoss): The loss function to optimize for. Defaults to
            ``optim.losses.Log``.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        intercept (float): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        clip_gradient (float): Clips the absolute value of each gradient value.
        initializer (optim.initializers.Initializer): Weights initialization scheme.

    Attributes:
        weights (collections.defaultdict): The current weights.

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing

            >>> X_y = datasets.Phishing()

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer=optim.SGD(.1))
            ... )

            >>> metric = metrics.Accuracy()

            >>> model_selection.progressive_val_score(X_y, model, metric)
            Accuracy: 88.96%

    Note:
        Using a feature scaler such as `preprocessing.StandardScaler` upstream helps the optimizer
        to converge.

    """

    def __init__(self, optimizer=None, loss=None, l2=.0, intercept=0., intercept_lr=.01,
                 clip_gradient=1e12, initializer=None):
        super().__init__(
            optimizer=optim.SGD(.01) if optimizer is None else optimizer,
            loss=optim.losses.Log() if loss is None else loss,
            intercept=intercept,
            intercept_lr=intercept_lr,
            l2=l2,
            clip_gradient=clip_gradient,
            initializer=initializer if initializer else optim.initializers.Zeros()
        )

    def predict_proba_one(self, x):
        p = self.loss.mean_func(self._raw_dot(x))  # Convert logit to probability
        return {False: 1. - p, True: p}



import random

from .. import base


class SEA(base.Dataset):
    """SEA generator.

    Each observation is composed of 3 features. Only the first two features are relevant. The
    target is binary, and is positive if the sum of the features exceeds a certain threshold. There
    are 4 thresholds to choose from. Concept drift can be introduced by switching the threshold
    anytime during the stream.

    Parameters:
        no (int): Determines the classification function to use. Possible choices are 0, 1, 2, 3.
        noise (float): Determines the amount of observations for which the target sign will be
            flipped.

    Example:

        ::

            >>> from creme import datasets

            >>> gen = datasets.gen.SEA(no=0, seed=42)

            >>> for x, y in gen.take(5):
            ...     print(x, y)
            {0: 6.394267984578837, 1: 0.25010755222666936, 2: 2.7502931836911926} False
            {0: 2.2321073814882277, 1: 7.364712141640124, 2: 6.766994874229113} True
            {0: 8.921795677048454, 1: 0.8693883262941615, 2: 4.2192181968527045} True
            {0: 0.29797219438070344, 1: 2.1863797480360336, 2: 5.053552881033624} False
            {0: 0.26535969683863625, 1: 1.988376506866485, 2: 6.498844377795232} False

    References:
        1. `A Streaming Ensemble Algorithm (SEA) for Large-Scale Classification <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.482.3991&rep=rep1&type=pdf>`_

    """

    def __init__(self, no, noise=0., seed=None):

        super().__init__(n_features=3, category=base.BINARY_CLF)

        if no not in (0, 1, 2, 3):
            raise ValueError('no must be one of (0, 1, 2, 3)')

        self.no = no
        self.noise = noise
        self.seed = seed
        self._threshold = {
            0: 8,
            1: 9,
            2: 7,
            3: 9.5
        }[no]
        self._rng = random.Random(seed)

    def __iter__(self):

        while True:

            x = {i: self._rng.uniform(0, 10) for i in range(3)}
            y = x[0] + x[1] > self._threshold

            if self.noise and self._rng.random() < self.noise:
                y = not y

            yield x, y



from .. import stream

from . import base


class HTTP(base.FileDataset):
    """HTTP dataset of the KDD 1999 cup.

    The goal is to predict whether or not an HTTP connection is anomalous or not. The dataset only
    contains 2,211 (0.4%) positive labels.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `HTTP (KDDCUP99) dataset <http://odds.cs.stonybrook.edu/http-kddcup99-dataset/>`_

    """

    def __init__(self, data_home=None, verbose=True):
        super().__init__(
            n_samples=567_498,
            n_features=3,
            category=base.BINARY_CLF,
            url='https://maxhalford.github.io/files/datasets/kdd99_http.zip',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            f'{directory}/kdd99_http.csv',
            target_name='service',
            converters={
                'duration': float,
                'src_bytes': float,
                'dst_bytes': float,
                'service': int
            }
        )



import surprise


class SurpriseWrapper(surprise.AlgoBase):

    def __init__(self, creme_recommender):
        self.creme_recommender = creme_recommender

    def fit(self, trainset):
        """Fits to an entire training set.

        This method is compatible with the API from the `surprise` library.

        """
        surprise.AlgoBase.fit(self, trainset)
        for user, item, y in trainset.all_ratings():
            self.creme_recommender.fit_one({'user': user, 'item': item}, y)
        return self

    def estimate(self, user, item):
        return self.creme_recommender.predict_one({'user': user, 'item': item})



import abc
import copy
import inspect
import sys
import types
import typing

from .. import utils


DEFAULT_TAGS = {
    'handles_text': False,
    'requires_positive_data': False,
    'handles_categorical_features': False
}


class Estimator(abc.ABC):
    """An estimator."""

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return _repr_obj(obj=self, params=self._get_params())

    def _get_params(self) -> typing.Dict[str, typing.Any]:
        return {
            name: getattr(self, name)
            for name, param in inspect.signature(self.__init__).parameters.items()  # type: ignore
            if param.kind != param.VAR_KEYWORD
        }

    def _set_params(self, new_params: typing.Optional[typing.Dict[str, typing.Any]] = None) -> 'Estimator':
        """Returns a new instance with the current parameters as well as new ones.

        The algorithm will be recursively called down ``Pipeline``s and ``TransformerUnion``s.

        Example:

            ::

                >>> from creme import linear_model
                >>> from creme import optim
                >>> from creme import preprocessing

                >>> model = (
                ...     preprocessing.StandardScaler() |
                ...     linear_model.LinearRegression(
                ...         optimizer=optim.SGD(lr=0.042),
                ...     )
                ... )

                >>> new_params = {
                ...     'LinearRegression': {
                ...         'l2': .001
                ...     }
                ... }

                >>> model._set_params(new_params)
                Pipeline (
                  StandardScaler (
                    with_mean=True
                    with_std=True
                  ),
                  LinearRegression (
                    optimizer=SGD (
                      lr=Constant (
                        learning_rate=0.042
                      )
                    )
                    loss=Squared ()
                    l2=0.001
                    intercept=0.
                    intercept_lr=Constant (
                      learning_rate=0.01
                    )
                    clip_gradient=1e+12
                    initializer=Zeros ()
                  )
                )

        """

        if new_params is None:
            new_params = {}

        params = {
            **self._get_params(),
            **new_params
        }

        return self.__class__(**copy.deepcopy(params))  # type: ignore

    @property
    def _tags(self) -> typing.Dict[str, bool]:
        """Returns the estimator's tags."""

        tags: typing.Dict[str, bool] = {}

        for base_class in inspect.getmro(self.__class__)[1:]:
            if isinstance(base_class, Estimator):
                _update_if_consistent(tags, base_class._more_tags)

        _update_if_consistent(tags, self._more_tags)

        return {**DEFAULT_TAGS, **tags}

    @property
    def _more_tags(self) -> typing.Dict[str, bool]:
        """Specific tags for this estimator."""
        return {}

    @property
    def _memory_usage(self) -> str:
        """Returns the memory usage in a human readable format."""

        def get_size(obj, seen=None):
            """Recursively finds size of objects"""
            size = sys.getsizeof(obj)
            if seen is None:
                seen = set()
            obj_id = id(obj)
            if obj_id in seen:
                return 0
            # Important mark as seen *before* entering recursion to gracefully handle
            # self-referential objects
            seen.add(obj_id)
            if isinstance(obj, dict):
                size += sum([get_size(v, seen) for v in obj.values()])
                size += sum([get_size(k, seen) for k in obj.keys()])
            elif hasattr(obj, '__dict__'):
                size += get_size(vars(obj), seen)
            elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
                size += sum([get_size(i, seen) for i in obj])
            return size

        mem_usage = get_size(self)
        return utils.pretty.humanize_bytes(mem_usage)


def _update_if_consistent(dict1: dict, dict2: dict):
    common_keys = set(dict1.keys()).intersection(dict2.keys())
    for key in common_keys:
        if dict1[key] != dict2[key]:
            raise TypeError(
                f'Inconsistent values for tag {key}: {dict1[key]} != {dict2[key]}')
    dict1.update(dict2)


def _repr_obj(obj: Estimator, params=None, show_modules: bool = False, depth: int = 0) -> str:
    """Returns a pretty representation of an instanted object."""

    rep = f'{obj.__class__.__name__} ('
    if show_modules:
        rep = f'{obj.__class__.__module__}.{rep}'
    tab = '\t'

    if params is None:
        params = {
            name: getattr(obj, name)
            for name, param in inspect.signature(obj.__init__).parameters.items()  # type: ignore
            if not (
                param.name == 'args' and param.kind == param.VAR_POSITIONAL or
                param.name == 'kwargs' and param.kind == param.VAR_KEYWORD
            )
        }

    n_params = 0

    for name, val in params.items():

        n_params += 1

        # Prettify the attribute when applicable
        if isinstance(val, types.FunctionType):
            val = val.__name__
        if isinstance(val, str):
            val = f'"{val}"'
        elif isinstance(val, float):
            val = (
                f'{val:.0e}'
                if (val > 1e5 or (val < 1e-4 and val > 0)) else
                f'{val:.6f}'.rstrip('0')
            )
        elif isinstance(val, set):
            val = sorted(val)
        elif hasattr(val, '__class__') and 'creme.' in str(type(val)):
            val = _repr_obj(obj=val, show_modules=show_modules, depth=depth + 1)

        rep += f'\n{tab * (depth + 1)}{name}={val}'

    if n_params:
        rep += f'\n{tab * depth}'
    rep += ')'

    return rep.expandtabs(2)



import itertools
import random

import numpy as np


def iter_array(X, y=None, feature_names=None, target_names=None, shuffle=False, seed=None):
    """Yields rows from an array of features and an array of targets.

    This method is compatible with ``numpy`` arrays as well as Python lists.

    Parameters:
        X (array-like of shape (n_samples, n_features))
        y (array-like of shape (n_samples,))
        feature_names (list of length n_features)
        target_names (list of length n_outputs)
        shuffle (bool): Whether to shuffle the inputs or not.
        seed (int): Random seed used for shuffling the data.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    feature_names = list(range(len(X[0]))) if feature_names is None else feature_names

    multioutput = y is not None and not np.isscalar(y[0])
    if multioutput and target_names is None:
        target_names = list(range(len(y[0])))

    # Shuffle the data
    rng = random.Random(seed)
    if shuffle:
        order = rng.sample(range(len(X)), k=len(X))
        X, y = X[order], y if y is None else y[order]

    if multioutput:

        for x, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
            yield dict(zip(feature_names, x)), dict(zip(target_names, yi))

    else:

        for x, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
            yield dict(zip(feature_names, x)), yi



def iter_pandas(X, y=None, **kwargs):
    """Yields rows from a ``pandas.DataFrame``.

    Parameters:
        X (pandas.DataFrame)
        y (array-like of shape (n_samples,))

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    import pandas as pd

    kwargs['feature_names'] = X.columns
    if isinstance(y, pd.DataFrame):
        kwargs['target_names'] = y.columns

    for x, yi in iter_array(X.to_numpy(), y, **kwargs):
        yield x, yi


def iter_sklearn_dataset(dataset, **kwargs):
    """Yields rows from one of the datasets provided by scikit-learn.

    Parameters:
        dataset (sklearn.utils.Bunch): A scikit-learn dataset.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    kwargs['X'] = dataset.data
    kwargs['y'] = dataset.target
    try:
        kwargs['feature_names'] = dataset.feature_names
    except AttributeError:
        pass

    for x, yi in iter_array(**kwargs):
        yield x, yi



"""Utilities for unit testing and sanity checking estimators."""
import copy
import functools
import math
import pickle


__all__ = [
    'check_estimator',
    'guess_model'
]


def guess_model(model):

    from .. import base
    from .. import compose

    if isinstance(model, compose.Pipeline):
        return guess_model(model.final_estimator)
    elif isinstance(model, base.Wrapper):
        return guess_model(model._model)
    return model


def yield_datasets(model):

    from .. import base
    from .. import datasets
    from .. import stream
    from sklearn import datasets as sk_datasets

    model = guess_model(model)

    if isinstance(model, (base.BinaryClassifier, base.MultiClassifier)):
        yield datasets.Phishing()
    if isinstance(model, base.MultiClassifier):
        yield datasets.ImageSegments().take(500)
    if isinstance(model, base.Regressor):
        yield datasets.TrumpApproval()
    if isinstance(model, base.MultiOutputRegressor):
        yield stream.iter_sklearn_dataset(sk_datasets.load_linnerud())
    if isinstance(model, base.MultiOutputRegressor):
        yield stream.iter_sklearn_dataset(sk_datasets.load_linnerud())


def check_fit_one(model, dataset):

    klass = model.__class__

    for x, y in dataset:

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.fit_one(x, y)

        # Check the model returns itself
        assert isinstance(model, klass)

        # Check fit_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one(classifier, dataset):

    for x, y in dataset:

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        classifier = classifier.fit_one(x, y)
        y_pred = classifier.predict_proba_one(x)

        # Check the probabilities are coherent
        assert isinstance(y_pred, dict)
        assert math.isclose(sum(y_pred.values()), 1.)
        for proba in y_pred.values():
            assert 0. <= proba <= 1.

        # Check predict_proba_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one_binary(classifier, dataset):

    for x, y in dataset:
        y_pred = classifier.predict_proba_one(x)
        classifier = classifier.fit_one(x, y)
        assert len(y_pred) == 2
        assert True in y_pred
        assert False in y_pred


def check_debug_one(model, dataset):
    for x, y in dataset:
        model.debug_one(x)
        model.fit_one(x, y)
        model.debug_one(x)
        break


def check_pickling(model, dataset):
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)
    for x, y in dataset:
        model.predict_one(x)
        model.fit_one(x, y)
    assert isinstance(pickle.loads(pickle.dumps(model)), model.__class__)


def check_repr_works(model):
    rep = repr(model)
    assert isinstance(rep, str)


def check_str_works(model):
    assert isinstance(str(model), str)


def check_tags(model):
    """Checks that the ``_tags`` property works."""
    assert isinstance(model._tags, dict)


def check_set_params_idempotent(model):
    assert len(model.__dict__) == len(model._set_params().__dict__)


def yield_checks(model):
    """Generates unit tests that can be applied to a given model.

    Parameters:
        model (base.Estimator)

    """

    from .. import base

    yield check_repr_works
    yield check_str_works
    yield check_tags
    yield check_set_params_idempotent

    for dataset in yield_datasets(model):

        def with_dataset(method):
            check = functools.partial(check_fit_one, dataset=dataset)
            functools.update_wrapper(check, method)
            return check

        yield with_dataset(check_fit_one)
        yield with_dataset(check_pickling)
        if hasattr(model, 'debug_one'):
            yield with_dataset(check_debug_one)

        model = guess_model(model)

        if isinstance(model, base.Classifier):
            yield with_dataset(check_predict_proba_one)

        if not isinstance(model, base.MultiClassifier):  # only binary classifiers
            yield with_dataset(check_predict_proba_one_binary)


def check_estimator(model):
    """Check if a model adheres to ``creme``'s conventions.

    Parameters:
        model (base.Estimator)

    """
    for check in yield_checks(model):
        check(copy.deepcopy(model))



import bisect
import collections
from copy import deepcopy


__all__ = ['simulate_qa']


class Memento(collections.namedtuple('Memento', 'i x y t_expire')):

    def __lt__(self, other):
        return self.t_expire < other.t_expire


def simulate_qa(X_y, moment, delay, copy=True):
    """Simulate a time-ordered question and answer session.

    Parameters:
        X_y (generator): A stream of (features, target) tuples.
        moment (callable or str): The attribute used for measuring time. If a callable
            is passed, then it is expected to take as input a `dict` of features. If ``None``, then
            the observations are implicitely timestamped in the order in which they arrive. If a
            ``str`` is passed, then it will be used to obtain the time from the input features.
        delay (callable or str or datetime.timedelta or int): The amount to wait before revealing
            the target associated with each observation to the model. This value is expected to be
            able to sum with the ``moment`` value. For instance, if ``moment`` is a
            `datetime.date`, then ``delay`` is expected to be a `datetime.timedelta`. If a callable
            is passed, then it is expected to take as input a `dict` of features and the target. If
            a `str` is passed, then it will be used to access the relevant field from the features.
            If ``None`` is passed, then no delay will be used, which leads to doing standard online
            validation. If a scalar is passed, such an `int` or a `datetime.timedelta`, then the
            delay is constant.
        copy (bool): If ``True``, then a separate copy of the features are yielded the second time
            around. This ensures that inadvertent modifications in downstream code don't have any
            effect.

    """

    # Determine how to insert mementos into the queue
    queue = lambda q, el: q.append(el)
    if callable(delay) or isinstance(delay, str):
        queue = lambda q, el: bisect.insort(q, el)

    # Coerce moment to a function
    if moment is None:
        get_moment = lambda i, _: i
    elif isinstance(moment, str):
        get_moment = lambda _, x: x[moment]
    elif callable(moment):
        get_moment = lambda _, x: moment(x)

    # Coerce delay to a function
    if delay is None:
        get_delay = lambda _, __: 0
    elif isinstance(delay, str):
        get_delay = lambda x, _: x[delay]
    elif not callable(delay):
        get_delay = lambda _, __: delay
    else:
        get_delay = delay

    mementos = []

    for i, (x, y) in enumerate(X_y):

        t = get_moment(i, x)
        d = get_delay(x, y)

        while mementos:

            # Get the oldest answer
            i_old, x_old, y_old, t_expire = mementos[0]

            # If the oldest answer isn't old enough then stop
            if t_expire > t:
                break

            # Reveal the duration and pop the trip from the queue
            yield i_old, x_old, y_old
            del mementos[0]

        queue(mementos, Memento(i, x, y, t + d))
        if copy:
            x = deepcopy(x)
        yield i, x, None

    for memento in mementos:
        yield memento.i, memento.x, memento.y



import os

from .. import stream

from . import base


class TrumpApproval(base.FileDataset):
    """Donald Trump approval ratings.

    This dataset was obtained by reshaping the data used by FiveThirtyEight for analyzing Donald
    Trump's approval ratings. It contains 5 features, which are approval ratings collected by
    5 polling agencies. The target is the approval rating from FiveThirtyEight's model. The goal of
    this task is to see if we can reproduce FiveThirtyEight's model.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Trump Approval Ratings <https://projects.fivethirtyeight.com/trump-approval-ratings/>`_

    """

    def __init__(self):
        super().__init__(
            n_samples=1001,
            n_features=6,
            category=base.REG
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            os.path.join(directory, 'trump_approval.csv.gz'),
            target_name='five_thirty_eight',
            converters={
                'ordinal_date': int,
                'gallup': float,
                'ipsos': float,
                'morning_consult': float,
                'rasmussen': float,
                'you_gov': float,
                'five_thirty_eight': float
            }
        )



from .. import stream

from . import base


class Bikes(base.FileDataset):
    """Bike sharing station information from the city of Toulouse.

    The goal is to predict the number of bikes in 5 different bike stations from the city of
    Toulouse.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `A short introduction and conclusion to the OpenBikes 2016 Challenge <https://maxhalford.github.io/blog/a-short-introduction-and-conclusion-to-the-openbikes-2016-challenge/>`_

    """

    def __init__(self, data_home=None, verbose=False):
        super().__init__(
            n_samples=182_470,
            n_features=8,
            category=base.REG,
            url='https://maxhalford.github.io/files/datasets/toulouse_bikes.zip',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            f'{directory}/toulouse_bikes.csv',
            target_name='bikes',
            converters={
                'clouds': int,
                'humidity': int,
                'pressure': float,
                'temperature': float,
                'wind': float,
                'bikes': int
            },
            parse_dates={'moment': '%Y-%m-%d %H:%M:%S'}
        )



import abc

from . import estimator


class AnomalyDetector(estimator.Estimator):

    @abc.abstractmethod
    def fit_one(self, x: dict) -> 'AnomalyDetector':
        """Updates the model."""

    @abc.abstractmethod
    def score_one(self, x: dict) -> float:
        """Returns an outlier score.

        The range of the score depends on each model. Some models will output anomaly scores
        between 0 and 1, others will not. In any case, the lower the score, the more likely it is
        that ``x`` is an anomaly.

        """



import os

from .. import stream

from . import base


class ChickWeights(base.FileDataset):
    """Chick weights along time.

    The stream contains 578 items and 3 features. The goal is to predict the weight of each chick
    along time, according to the diet the chick is on. The data is ordered by time and then by
    chick.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Chick weight dataset overview <http://rstudio-pubs-static.s3.amazonaws.com/107631_131ad1c022df4f90aa2d214a5c5609b2.html>`_

    """

    def __init__(self):
        super().__init__(
            n_samples=578,
            n_features=3,
            category=base.REG
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            os.path.join(directory, 'chick-weights.csv'),
            target_name='weight',
            converters={'time': int, 'weight': int, 'chick': int, 'diet': int}
        )



import numbers
import functools
import collections

from .. import base
from .. import stats
from .. import utils


__all__ = [
    'Binarizer',
    'MaxAbsScaler',
    'MinMaxScaler',
    'Normalizer',
    'RobustScaler',
    'StandardScaler'
]


def safe_div(a, b):
    if b == 0:
        return a
    return a / b


class Binarizer(base.Transformer):
    """Binarizes the data to 0 or 1 according to a threshold.

    Parameters:
        threshold (float): Values above this are replaced by 1 and remaining by 0. Defaults to ``0.0``.

    Attributes:
        threshold (float): Threshold value for binarizing.

    Example:

        ::

              >>> import creme
              >>> import numpy as np

              >>> rng = np.random.RandomState(42)
              >>> X = [{'x1': v, 'x2': int(v)} for v in rng.uniform(low=-4, high=4, size=6)]

              >>> binarizer = creme.preprocessing.Binarizer()
              >>> for x in X:
              ...     print(binarizer.fit_one(x).transform_one(x))
              {'x1': 0.0, 'x2': 0}
              {'x1': 1.0, 'x2': 1}
              {'x1': 1.0, 'x2': 1}
              {'x1': 1.0, 'x2': 0}
              {'x1': 0.0, 'x2': 0}
              {'x1': 0.0, 'x2': 0}
    """

    def __init__(self, threshold=0.0):
        self.threshold = threshold

    def transform_one(self, x):
        x_tf = x.copy()

        for i, xi in x_tf.items():
            if isinstance(xi, numbers.Number):
                x_tf[i] = type(xi)(xi > self.threshold)

        return x_tf


class StandardScaler(base.Transformer):
    """Scales the data so that it has zero mean and unit variance.

    Under the hood a running mean and a running variance are maintained. The scaling is slightly
    different than when using scikit-learn but this doesn't seem to have any impact on learning
    performance.

    Parameters:
        with_mean (bool): Whether to centre the data before scaling. Defaults to ``True``.
        with_std (bool): Whether to scale data. Defaults to ``True``.

    Attributes:
        variances (dict): Mapping between features and instances of `stats.Var`.

    Example:

        ::

              >>> import creme
              >>> import numpy as np
              >>> from sklearn import preprocessing

              >>> rng = np.random.RandomState(42)
              >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

              >>> scaler = creme.preprocessing.StandardScaler()
              >>> for x in X:
              ...     print(scaler.fit_one(x).transform_one(x))
              {'x': 0.0}
              {'x': 0.7071067811865474}
              {'x': 0.15899361505958234}
              {'x': -0.2705322151649623}
              {'x': -1.3161741269826}
              {'x': -1.051218830939811}
              {'x': -1.1096762396286515}
              {'x': 1.09141268517007}
              {'x': 0.3109060850298578}
              {'x': 0.5949866915752465}
              {'x': -1.3540960661327461}
              {'x': 1.2959176347038681}
              {'x': 0.8426620966002304}
              {'x': -0.8843412116857524}
              {'x': -0.91188180555936}

              >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
              >>> preprocessing.StandardScaler().fit_transform(X)
              array([[-0.36224883],
                     [ 1.37671717],
                     [ 0.71659166],
                     [ 0.31416852],
                     [-1.02177407],
                     [-1.02184687],
                     [-1.31735428],
                     [ 1.1215704 ],
                     [ 0.32158263],
                     [ 0.64439399],
                     [-1.43053132],
                     [ 1.43465174],
                     [ 1.01975844],
                     [-0.85179183],
                     [-0.94388734]])

    """

    def __init__(self, with_mean=True, with_std=True):
        self.with_mean = with_mean
        self.with_std = with_std
        self.variances = collections.defaultdict(stats.Var)

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.variances[i].update(xi)

        return self

    def transform_one(self, x):
        x_tf = {}

        for i, xi in x.items():
            x_tf[i] = xi
            if self.with_mean:
                x_tf[i] -= self.variances[i].mean.get()
            if self.with_std:
                x_tf[i] = safe_div(x_tf[i], self.variances[i].get() ** .5)

        return x_tf


class MinMaxScaler(base.Transformer):
    """Scales the data to a fixed range from 0 to 1.

    Under the hood a running min and a running peak to peak (max - min) are maintained. The scaling
    is slightly different than when using scikit-learn but this doesn't seem to have any impact on
    learning performance.

    Attributes:
        min (dict): Mapping between features and instances of `stats.Min`.
        max (dict): Mapping between features and instances of `stats.Max`.

    Example:

        ::

            >>> import creme
            >>> import numpy as np
            >>> from sklearn import preprocessing

            >>> rng = np.random.RandomState(42)
            >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

            >>> scaler = creme.preprocessing.MinMaxScaler()
            >>> for x in X:
            ...     print(scaler.fit_one(x).transform_one(x))
            {'x': 0.0}
            {'x': 1.0}
            {'x': 0.6203919416734277}
            {'x': 0.3889767542308411}
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.9052932403284701}
            {'x': 0.6083494586037179}
            {'x': 0.7281723223510779}
            {'x': 0.0}
            {'x': 1.0}
            {'x': 0.8551948389216542}
            {'x': 0.20199040802352813}
            {'x': 0.16984743067826635}

            >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
            >>> preprocessing.MinMaxScaler().fit_transform(X)
            array([[0.37284965],
                   [0.9797798 ],
                   [0.74938422],
                   [0.60893137],
                   [0.14266357],
                   [0.14263816],
                   [0.03950081],
                   [0.89072903],
                   [0.61151903],
                   [0.72418595],
                   [0.        ],
                   [1.        ],
                   [0.85519484],
                   [0.20199041],
                   [0.16984743]])

    """

    def __init__(self):
        self.min = collections.defaultdict(stats.Min)
        self.max = collections.defaultdict(stats.Max)

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.min[i].update(xi)
            self.max[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi - self.min[i].get(),
                        self.max[i].get() - self.min[i].get())
            for i, xi in x.items()
        }


class MaxAbsScaler(base.Transformer):
    """Scales the data to a [-1, 1] range based on absolute maximum.

    Under the hood a running absolute max is maintained. This scaler is meant for
    data that is already centered at zero or sparse data. It does not shift/center
    the data, and thus does not destroy any sparsity.

    Attributes:
        abs_max (dict): Mapping between features and instances of `stats.AbsMax`.

    Example:

        ::

            >>> import creme
            >>> import numpy as np
            >>> from sklearn import preprocessing

            >>> rng = np.random.RandomState(42)
            >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

            >>> scaler = creme.preprocessing.MaxAbsScaler()
            >>> for x in X:
            ...     print(scaler.fit_one(x).transform_one(x))
            {'x': 1.0}
            {'x': 1.0}
            {'x': 0.9258754518784218}
            {'x': 0.8806879332749703}
            {'x': 0.7306768519605097}
            {'x': 0.7306686776326253}
            {'x': 0.6974865739110592}
            {'x': 0.9713499336579836}
            {'x': 0.8815204528926222}
            {'x': 0.9177684779286244}
            {'x': 0.6847780857355226}
            {'x': 1.0}
            {'x': 0.9537133387191868}
            {'x': 0.7449179338112799}
            {'x': 0.734643499572489}

            >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
            >>> preprocessing.MaxAbsScaler().fit_transform(X)
            array([[0.79953273],
                   [0.99353666],
                   [0.9198912 ],
                   [0.87499575],
                   [0.72595424],
                   [0.72594612],
                   [0.69297848],
                   [0.96507177],
                   [0.87582288],
                   [0.91183663],
                   [0.68035213],
                   [1.        ],
                   [0.95371334],
                   [0.74491793],
                   [0.7346435 ]])
    """

    def __init__(self):
        self.abs_max = collections.defaultdict(stats.AbsMax)

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.abs_max[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: safe_div(xi, self.abs_max[i].get())
            for i, xi in x.items()
        }


class RobustScaler(base.Transformer):
    """Scale features using statistics that are robust to outliers.

    This Scaler removes the median and scales the data according to the
    interquantile range.

    Parameters:
        with_centering (bool): Whether to centre the data before scaling. Defaults to ``True``.
        with_scaling (bool): Whether to scale data to iqr. Defaults to ``True``.
        q_inf (float): Desired inferior quantile, must be between 0 and 1. Defaults to ``0.25``.
        q_sup (float): Desired superior quantile, must be between 0 and 1. Defaults to ``0.75``.

    Attributes:
        median (dict): Mapping between features and instances of ``stats.Quantile(0.5)``.
        iqr (dict): Mapping between features and instances of `stats.IQR`.

    Example:

        ::

            >>> import creme
            >>> import numpy as np
            >>> from sklearn import preprocessing

            >>> rng = np.random.RandomState(42)
            >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

            >>> scaler = creme.preprocessing.RobustScaler()
            >>> for x in X:
            ...     print(scaler.fit_one(x).transform_one(x))
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.0}
            {'x': -0.3787338518541633}
            {'x': -1.2383133577483856}
            {'x': -2.6296427694340427}
            {'x': -1.1145387387139178}
            {'x': 1.3429672438986109}
            {'x': 0.037059676555904184}
            {'x': 0.25865683357370733}
            {'x': -1.165696864901573}
            {'x': 0.7074284260047643}
            {'x': 0.25014164136194217}
            {'x': -0.8213524413984991}
            {'x': -0.6509250023240815}

            >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
            >>> preprocessing.RobustScaler().fit_transform(X)
            array([[-0.36543233],
                   [ 0.57403854],
                   [ 0.21740783],
                   [ 0.        ],
                   [-0.72173876],
                   [-0.72177808],
                   [-0.88142503],
                   [ 0.4361963 ],
                   [ 0.00400545],
                   [ 0.17840326],
                   [-0.94256856],
                   [ 0.60533751],
                   [ 0.38119272],
                   [-0.62990639],
                   [-0.6796607 ]])
    """

    def __init__(self, with_centering=True, with_scaling=True, q_inf=0.25, q_sup=0.75):
        self.with_centering = with_centering
        self.with_scaling = with_scaling
        self.q_inf = q_inf
        self.q_sup = q_sup
        self.median = collections.defaultdict(
            functools.partial(stats.Quantile, 0.5)
        )
        self.iqr = collections.defaultdict(
            functools.partial(stats.IQR, self.q_inf, self.q_sup)
        )

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            if self.with_centering:
                self.median[i].update(xi)
            if self.with_scaling:
                self.iqr[i].update(xi)

        return self

    def transform_one(self, x):
        x_tf = {}

        for i, xi in x.items():
            x_tf[i] = xi
            if self.with_centering:
                x_tf[i] -= self.median[i].get()
            if self.with_scaling:
                x_tf[i] = safe_div(x_tf[i], self.iqr[i].get())

        return x_tf


class Normalizer(base.Transformer):
    """Scales a set of features so that it has unit norm.

    This is particularly useful when used after a `feature_extraction.TFIDF`.

    Parameters:
        order (int): Order of the norm (e.g. 2 corresponds to the $L^2$ norm).

    Example:

        ::

            >>> from creme import preprocessing
            >>> from creme import stream

            >>> scaler = preprocessing.Normalizer(order=2)

            >>> X = [[4, 1, 2, 2],
            ...      [1, 3, 9, 3],
            ...      [5, 7, 5, 1]]

            >>> for x, _ in stream.iter_array(X):
            ...     print(scaler.transform_one(x))
            {0: 0.8, 1: 0.2, 2: 0.4, 3: 0.4}
            {0: 0.1, 1: 0.3, 2: 0.9, 3: 0.3}
            {0: 0.5, 1: 0.7, 2: 0.5, 3: 0.1}

    """

    def __init__(self, order=2):
        self.order = order

    def transform_one(self, x):
        norm = utils.math.norm(x, order=self.order)
        return {i: xi / norm for i, xi in x.items()}



import os

from .. import stream

from . import base


class Bananas(base.FileDataset):
    """Bananas dataset.

    An artificial dataset where instances belongs to several clusters with a banana shape.
    There are two attributes that correspond to the x and y axis, respectively.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `OpenML page <https://www.openml.org/d/1460>`_

    """

    def __init__(self):
        super().__init__(
            n_samples=5300,
            n_features=2,
            category=base.BINARY_CLF
        )

    def _stream_X_y(self, directory):
        return stream.iter_libsvm(
            os.path.join(directory, 'banana.zip'),
            target_type=lambda x: x == '1'
        )



import itertools
import os
import shutil
import tarfile
from urllib import request
import zipfile


REG = 'Regression'
BINARY_CLF = 'Binary'
MULTI_CLF = 'Multiclass'


def get_data_home(data_home=None):
    """Return the path of the creme data directory.

    By default this will expand the relative path '~/creme_data'.

    """
    if data_home is None:
        data_home = os.environ.get('CREME_DATA', os.path.join('~', 'creme_data'))
        data_home = os.path.expanduser(data_home)
        if not os.path.exists(data_home):
            os.makedirs(data_home)
    return data_home


def download_dataset(url, data_home, uncompress=True, verbose=True):
    """Downloads/decompresses a dataset locally if does not exist.

    Parameters:
        url (str): From where to download the dataset.
        data_home (str): The directory where you wish to store the data.
        uncompress (bool): Whether to uncompress the file or not.
        verbose (bool): Whether to indicate download progress or not.

    Returns:
        data_dir_path (str): Where the dataset is stored.

    """

    def _print(msg):
        if verbose:
            print(msg)

    name = os.path.basename(url)
    extension = '.'.join(name.split('.')[1:])
    data_home = get_data_home(data_home=data_home)
    path = os.path.join(data_home, f'{name}')
    archive_path = path
    if extension:
        path = path[:-(len(extension) + 1)]  # e.g. path/to/file.tar.gz becomes path/to/file

    # Download if necessary
    if not (os.path.exists(path) or os.path.exists(archive_path)):

        _print(f'Downloading {url}')
        with request.urlopen(url) as r, open(archive_path, 'wb') as f:
            shutil.copyfileobj(r, f)

    # If no uncompression is required then we're done
    if not uncompress:
        return archive_path

    # Uncompress if necessary
    if not os.path.exists(path):

        _print(f'Uncompressing into {path}')

        if extension.endswith('zip'):
            with zipfile.ZipFile(archive_path, 'r') as zf:
                zf.extractall(path)

        elif extension.endswith(('gz', 'tar')):
            mode = 'r:' if extension.endswith('tar') else 'r:gz'
            tar = tarfile.open(archive_path, mode)
            tar.extractall(path)
            tar.close()

        else:
            raise RuntimeError(f'Unhandled extension type: {extension}')

        # Delete the archive file now that the dataset is available
        os.remove(archive_path)

    return path


class Dataset:

    def __init__(self, n_features, category):
        self.n_features = n_features
        self.category = category

    def __iter__(self):
        raise NotImplementedError

    def take(self, k):
        """Yields the k first (``x``, ``y``) pairs."""
        return itertools.islice(self, k)


class FileDataset(Dataset):

    def __init__(self, n_samples, n_features, category, **dl_params):
        super().__init__(n_features=n_features, category=category)
        self.n_samples = n_samples
        self.dl_params = dl_params

    def _stream_X_y(self, dir):
        raise NotImplementedError

    @property
    def _remote(self):
        """Whether or not the dataset needs downloading or not."""
        return 'url' in self.dl_params

    @property
    def _ready(self):
        """Whether or not the dataset is ready to be read."""
        if self._remote:
            return os.path.isdir(os.path.join(
                get_data_home(self.dl_params['data_home']),
                self.dl_params['name']
            ))
        return True

    def __iter__(self):
        if self._remote:
            data_dir_path = download_dataset(**self.dl_params)
        else:
            data_dir_path = os.path.dirname(__file__)
        yield from self._stream_X_y(data_dir_path)



import ast

from .. import stream

from . import base


class Restaurants(base.FileDataset):
    """Data from the Kaggle Recruit Restaurants challenge.

    The goal is to predict the number of visitors in each of 829 Japanese restaurants over a priod
    of roughly 16 weeks. The data is ordered by date and then by restaurant ID.

    Parameters:
        data_home (str): The directory where you wish to store the data.
        verbose (bool): Whether to indicate download progress or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `Recruit Restaurant Visitor Forecasting <https://www.kaggle.com/c/recruit-restaurant-visitor-forecasting>`_

    """

    def __init__(self, data_home=None, verbose=True):
        super().__init__(
            n_samples=252_108,
            n_features=7,
            category=base.REG,
            url='https://maxhalford.github.io/files/datasets/kaggle_recruit_restaurants.zip',
            data_home=data_home,
            verbose=verbose
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            f'{directory}/kaggle_recruit_restaurants.csv',
            target_name='visitors',
            converters={
                'latitude': float,
                'longitude': float,
                'visitors': int,
                'is_holiday': ast.literal_eval
            },
            parse_dates={'date': '%Y-%m-%d'}
        )



import os

from .. import stream

from . import base


class Phishing(base.FileDataset):
    """Phishing websites.

    This dataset contains features from web pages that are classified as phishing or not.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    References:
        1. `UCI page <http://archive.ics.uci.edu/ml/datasets/Website+Phishing>`_

    """

    def __init__(self):
        super().__init__(
            n_samples=1250,
            n_features=9,
            category=base.BINARY_CLF
        )

    def _stream_X_y(self, directory):
        return stream.iter_csv(
            os.path.join(directory, 'phishing.csv.gz'),
            target_name='is_phishing',
            converters={
                'empty_server_form_handler': float,
                'popup_window': float,
                'https': float,
                'request_from_other_domain': float,
                'anchor_from_other_domain': float,
                'is_popular': float,
                'long_url': float,
                'age_of_domain': int,
                'ip_in_url': int,
                'is_phishing': lambda x: x == '1'
            }
        )



import collections
import math
import random

from creme import base


class RBFSampler(base.Transformer):
    """Extracts random features which approximate an RBF kernel.

    This is a powerful way to give non-linear capacity to linear classifiers. In the litterature
    this is also called "random Fourier features".

    Parameters:
        gamma (float): RBF kernel parameter: ``(-gamma * x^2)``.
        n_components (int) Number of Monte Carlo samples per original feature. Equals the
            dimensionality of the computed feature space.
        seed (int): Random number seed.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream

            >>> # XOR function
            >>> X = [[0, 0], [1, 1], [1, 0], [0, 1]]
            >>> Y = [0, 0, 1, 1]

            >>> model = linear_model.LogisticRegression(optimizer=optim.SGD(0.1))

            >>> for x, y in stream.iter_array(X, Y):
            ...     model = model.fit_one(x, y)
            ...     y_pred = model.predict_one(x)
            ...     print(y, int(y_pred))
            0 0
            0 0
            1 0
            1 1

            >>> model = (
            ...     preprocessing.RBFSampler(seed=3) |
            ...     linear_model.LogisticRegression(optimizer=optim.SGD(0.1))
            ... )

            >>> for x, y in stream.iter_array(X, Y):
            ...     model = model.fit_one(x, y)
            ...     y_pred = model.predict_one(x)
            ...     print(y, int(y_pred))
            0 0
            0 0
            1 1
            1 1

    References:
        1. `Rahimi, A. and Recht, B., 2008. Random features for large-scale kernel machines. In Advances in neural information processing systems (pp. 1177-1184). <https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf>`_

    """

    def __init__(self, gamma=1., n_components=100, seed=None):
        self.gamma = gamma
        self.n_components = n_components
        self.seed = seed
        self.rng = random.Random(seed)
        self.weights = collections.defaultdict(self._random_weights)
        self.offsets = [random.uniform(0, 2 * math.pi)
                        for _ in range(n_components)]

    def _random_weights(self):
        return [
            math.sqrt(2 * self.gamma) * self.rng.gauss(mu=0, sigma=1)
            for _ in range(self.n_components)
        ]

    def transform_one(self, x, y=None):
        return {
            (i, j): math.cos(xi * wj + self.offsets[j])
            for i, xi in x.items()
            for j, wj in enumerate(self.weights[i])
        }



from .. import base
from ..utils import math


__all__ = ['PredClipper']


class PredClipper(base.Regressor, base.Wrapper):
    """Clips the target after predicting.

    Parameters:
        regressor (creme.base.Regressor): regressor model applied for training.
        y_min (float): minimum value.
        y_max (float): maximum value.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import meta

            >>> X_y = (
            ...     ({'a': 2, 'b': 4}, 80),
            ...     ({'a': 3, 'b': 5}, 100),
            ...     ({'a': 4, 'b': 6}, 120)
            ... )

            >>> model = meta.PredClipper(
            ...     regressor=linear_model.LinearRegression(),
            ...     y_min=0,
            ...     y_max=200
            ... )

            >>> for x, y in X_y:
            ...     _ = model.fit_one(x, y)

            >>> model.predict_one({'a': -100, 'b': -200})
            0

            >>> model.predict_one({'a': 50, 'b': 60})
            200

    """

    def __init__(self, regressor, y_min, y_max):
        self.regressor = regressor
        self.y_min = y_min
        self.y_max = y_max

    @property
    def _model(self):
        return self.regressor

    @property
    def _labelloc(self):
        return 'b'  # for bottom

    def fit_one(self, x, y):
        self.regressor.fit_one(x, y)
        return self

    def predict_one(self, x):
        y_pred = self.regressor.predict_one(x)
        return math.clamp(y_pred, self.y_min, self.y_max)



import abc
import typing

from . import estimator


class MultiOutputEstimator(estimator.Estimator):
    """A multi-output estimator."""


class MultiOutputClassifier(MultiOutputEstimator):
    """A multi-output classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: typing.Dict[typing.Hashable, typing.Hashable]) -> 'MultiOutputClassifier':
        """Fits to a set of features ``x`` and a set of labels ``y``.

        Parameters:
            x (dict)
            y (dict of Label)

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_proba_one(self, x: dict) -> typing.Dict[typing.Hashable, typing.Dict[typing.Hashable, float]]:
        """Given a set of features ``x``, predicts a the probability of each label for each output.

        Parameters:
            x (dict)

        Returns:

        """

    def predict_one(self, x: dict) -> typing.Dict[typing.Hashable, typing.Hashable]:
        """Given a set of features ``x``, predicts a label for each output.

        Parameters:
            x (dict)

        Returns:

        """
        y_pred = self.predict_proba_one(x)
        return {
            c: max(y_pred[c], key=y_pred[c].get)
            for c in y_pred
        }


class MultiOutputRegressor(MultiOutputEstimator):
    """A multi-output regressor."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: typing.Dict[typing.Hashable, float]) -> 'MultiOutputRegressor':
        """Fits to a set of features ``x`` and a set of outputs ``y``.

        Parameters:
            x (dict)
            y (dict of Label)

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> typing.Dict[typing.Hashable, float]:
        """Given a set of features ``x``, predicts a label for each output.

        Parameters:
            x (dict)

        Returns:
            dict

        """



import abc
import typing

from . import estimator


class Regressor(estimator.Estimator):
    """A regressor."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: typing.Union[float, int]) -> 'Regressor':
        """Fits to a set of features ``x`` and a real-valued target ``y``.

        Parameters:
            x (dict)
            y (float)

        Returns:
            self: object

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> float:
        """Predicts the target value of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            float: The prediction.

        """
